"""Utilities for models."""
from __future__ import annotations

from functools import cached_property
from typing import Annotated

import numpy as np
import pandas as pd
from pydantic import BaseModel, ConfigDict, PlainSerializer, PlainValidator
from typing_extensions import Self

from dascore.compat import array
from dascore.units import Quantity, get_quantity, get_quantity_str
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import (
    all_close,
    to_str,
    unbyte,
)
from dascore.utils.time import to_datetime64, to_timedelta64

# --- A list of custom types with appropriate serialization/deserialization
# these can just be use with pydantic type-hints.

frozen_dict_validator = PlainValidator(lambda x: FrozenDict(x))
frozen_dict_serializer = PlainSerializer(lambda x: dict(x))

# A datetime64
DateTime64 = Annotated[
    np.datetime64,
    PlainValidator(to_datetime64),
    PlainSerializer(to_str, when_used="json"),  # getting undefined name
]

TimeDelta64 = Annotated[
    np.timedelta64,
    PlainValidator(to_timedelta64),
    PlainSerializer(to_str, when_used="json"),  # getting undefined name
]

ArrayLike = Annotated[
    np.ndarray,
    PlainValidator(array),
]

DTypeLike = Annotated[
    str,
    PlainValidator(np.dtype),
]

UnitQuantity = Annotated[
    Quantity | str | None,
    PlainValidator(get_quantity),
    PlainSerializer(get_quantity_str),
]

CommaSeparatedStr = Annotated[
    str, PlainValidator(lambda x: x if isinstance(x, str) else ",".join(x))
]

FrozenDictType = Annotated[
    FrozenDict,
    frozen_dict_validator,
    frozen_dict_serializer,
]

UTF8Str = Annotated[str, PlainValidator(unbyte)]


def sensible_model_equals(self, other):
    """Custom equality to not compare private attrs and handle numpy arrays."""
    try:
        d1, d2 = self.model_dump(), other.model_dump()
    except (TypeError, ValueError, AttributeError):
        return False
    if not set(d1) == set(d2):  # different keys, not equal
        return False
    for name, val1 in d1.items():
        # skip any private attributes.
        if name.startswith("_"):
            continue
        val2 = d2[name]
        if isinstance(val1, np.ndarray):
            if not all_close(val1, val2):
                return False
        else:
            if not val1 == val2:
                return False
    return True


class DascoreBaseModel(BaseModel):
    """A base model with sensible configurations."""

    _cache = {}

    model_config = ConfigDict(
        extra="ignore",  # TODO: change to raise, then let subclass overwrite
        validate_assignment=True,
        ignored_types=(cached_property,),
        frozen=True,
        validate_default=True,
        arbitrary_types_allowed=True,
    )

    def new(self, **kwargs) -> Self:
        """Create new instance with some attributed updated."""
        out = dict(self)
        for item, value in kwargs.items():
            out[item] = value
        return self.__class__(**out)

    @classmethod
    def get_summary_df(cls):
        """Get dataframe of attributes and descriptions for display."""
        fields = cls.model_fields
        names_desc = {
            i: v.description
            for i, v in fields.items()
            if getattr(v, "description", False)
        }
        out = pd.Series(names_desc).to_frame(name="description")
        out.index.name = "attribute"
        return out

    __eq__ = sensible_model_equals



"""Module for wiggle plotting."""
from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np

from dascore.constants import PatchType
from dascore.utils.patch import patch_function
from dascore.utils.plotting import (
    _format_time_axis,
    _get_ax,
    _get_dim_label,
)


def _get_offsets_factor(patch, dim, scale, other_labels):
    """Get the offsets and scale the data."""
    dim_axis = patch.dims.index(dim)
    # get and apply scale_factor. This controls how far apart the wiggles are.
    diffs = np.max(patch.data, axis=dim_axis) - np.min(patch.data, axis=dim_axis)
    offsets = (np.median(diffs) * scale) * np.arange(len(other_labels))
    # add scale factor to data
    data_scaled = offsets[None, :] + patch.data
    return offsets, data_scaled


def _shade(offsets, ax, data_scaled, color, wiggle_labels):
    """Shades the part of the waveforms under the offset line."""
    for i in range(len(offsets)):
        ax.fill_between(
            wiggle_labels,
            offsets[i],
            data_scaled[:, i],
            where=(data_scaled[:, i] > offsets[i]),
            color=color,
            alpha=0.6,
        )


def _format_y_axis_ticks(ax, offsets, other_axis_ticks, max_ticks=10):
    """Format the Y axis tick labels."""
    # set the offset
    ax.set_yticks(offsets, other_axis_ticks)
    min_bins = min(len(other_axis_ticks), max_ticks)
    plt.locator_params(axis="y", nbins=min_bins)


@patch_function()
def wiggle(
    patch: PatchType,
    dim="time",
    scale=1,
    alpha=0.2,
    color="black",
    shade=False,
    ax: plt.Axes | None = None,
    show=False,
) -> plt.Figure:
    """
    Create a wiggle plot of patch data.

    Parameters
    ----------
    patch
        The Patch object.
    dim
        The dimension along which samples are connected.
    scale
        The scale (or gain) of the waveforms. A value of 1 indicates waveform
        centroids are separated by the average total waveform excursion.
    alpha
        Opacity of the wiggle lines.
    color
        Color of wiggles
    shade
        If True, shade all values of each trace which are less than the mean
        trace value.
    ax
        A matplotlib object, if None ne will be created.
    show
        If True, show the plot, else just return axis.

    Examples
    --------
    >>> # Plot the default patch
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> _ = patch.viz.wiggle()
    """
    ax = _get_ax(ax)
    assert len(patch.dims) == 2, "Can only make wiggle plot of 2D Patch"
    # After transpose selected dim must be axis 0 and other axis 1
    patch = patch.transpose(dim, ...)
    other_dim = next(iter(set(patch.dims) - {dim}))
    # values for axis which is connected
    connect_axis_ticks = patch.coords.get_array(dim)
    # values for y axis (not connected)
    other_axis_ticks = patch.coords.get_array(other_dim)
    offsets, data_scaled = _get_offsets_factor(patch, dim, scale, other_axis_ticks)
    # now plot, add labels, etc.
    ax.plot(connect_axis_ticks, data_scaled, color=color, alpha=alpha)
    # shade negative part of waveforms if desired
    if shade:
        _shade(offsets, ax, data_scaled, color, connect_axis_ticks)
    _format_y_axis_ticks(ax, offsets, other_axis_ticks)
    for dim, x in zip(patch.dims, ["x", "y"]):
        getattr(ax, f"set_{x}label")(_get_dim_label(patch, dim))
        # format all dims which have time types.
        if np.issubdtype(patch.get_coord(dim).dtype, np.datetime64):
            _format_time_axis(ax, dim, x)
    if show:
        plt.show()
    ax.invert_yaxis()  # invert y so its consistent with waterfall
    return ax



"""Processing for applying a taper."""
from __future__ import annotations

from collections.abc import Sequence

import numpy as np
from scipy.signal import windows  # the best operating system?

from dascore.constants import PatchType
from dascore.exceptions import ParameterError
from dascore.units import Quantity
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import broadcast_for_index
from dascore.utils.patch import get_dim_value_from_kwargs, patch_function

TAPER_FUNCTIONS = dict(
    barthann=windows.barthann,
    bartlett=windows.bartlett,
    blackman=windows.blackman,
    blackmanharris=windows.blackmanharris,
    bohman=windows.bohman,
    hamming=windows.hamming,
    hann=windows.hann,
    nuttall=windows.nuttall,
    parzen=windows.parzen,
    triang=windows.triang,
)


def _get_taper_slices(patch, kwargs):
    """Get slice for start/end of patch."""
    dim, axis, value = get_dim_value_from_kwargs(patch, kwargs)
    coord = patch.coords.coord_map[dim]
    if isinstance(value, Sequence | np.ndarray):
        assert len(value) == 2, "Length 2 sequence required."
        start, stop = value[0], value[1]
    else:
        start, stop = value, value
    dur = coord.max() - coord.min()
    # either let units pass through or multiply by d_len
    start = start if isinstance(start, Quantity) or start is None else start * dur
    stop = stop if isinstance(stop, Quantity) or stop is None else stop * dur
    stop = -stop if stop is not None else stop
    _, inds_1 = coord.select((None, start), relative=True)
    _, inds_2 = coord.select((stop, None), relative=True)
    return axis, (start, stop), inds_1, inds_2


def _get_window_function(window_type):
    """Get the window function to use for taper."""
    # get taper function or raise if it isn't known.
    if window_type not in TAPER_FUNCTIONS:
        msg = (
            f"'{window_type}' is not a known window type. "
            f"Options are: {sorted(TAPER_FUNCTIONS)}"
        )
        raise ParameterError(msg)
    func = TAPER_FUNCTIONS[window_type]
    return func


def _validate_windows(samps, start_slice, end_slice, shape, axis):
    """Validate the the windows don't overlap or exceed dim len."""
    max_len = shape[axis]
    start_ind = start_slice.stop
    end_ind = end_slice.start

    bad_start = samps[0] is not None and (start_ind is None or start_ind < 0)
    bad_end = samps[1] is not None and (end_ind is None or end_ind > max_len)

    if bad_start or bad_end:
        msg = "Total taper lengths exceed total dim length"
        raise ParameterError(msg)

    if start_ind is None or end_ind is None:
        return
    if start_ind > end_ind:
        msg = "Taper windows cannot overlap"
        raise ParameterError(msg)


@patch_function()
@compose_docstring(taper_type=sorted(TAPER_FUNCTIONS))
def taper(
    patch: PatchType,
    window_type: str = "hann",
    **kwargs,
) -> PatchType:
    """
    Taper the ends of the signal.

    Parameters
    ----------
    patch
        The patch instance.
    window_type
        The type of window to use For tapering. Supported Options are:
            {taper_type}.
    **kwargs
        Used to specify the dimension along which to taper and the percentage
        of total length of the dimension or abolsute units. If a single value
        is passed, the taper will be applied to both ends. A length two tuple
        can specify different values for each end, or no taper on one end.

    Returns
    -------
    The tapered patch.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch() # generate example patch
    >>> # Apply an Hanning taper to 5% of each end for time dimension.
    >>> patch_taper1 = patch.taper(time=0.05, window_type="hann")
    >>> # Apply a triangular taper to 10% of the start of the distance dimension.
    >>> patch_taper2 = patch.taper(distance=(0.10, None), window_type='triang')
    >>> # Apply taper on first and last 15 m along distance axis.
    >>> from dascore.units import m
    >>> patch_taper3 = patch.taper(distance=15 * m)
    """
    func = _get_window_function(window_type)
    # get taper values in samples.
    out = np.array(patch.data)
    shape = out.shape
    n_dims = len(out.shape)
    axis, samps, start_slice, end_slice = _get_taper_slices(patch, kwargs)
    _validate_windows(samps, start_slice, end_slice, shape, axis)
    if samps[0] is not None:
        val = start_slice.stop
        window = func(2 * val)[:val]
        # get indices window (which will broadcast) and data
        data_inds = broadcast_for_index(n_dims, axis, start_slice)
        window_inds = broadcast_for_index(n_dims, axis, slice(None), fill=None)
        out[data_inds] = out[data_inds] * window[window_inds]
    if samps[1] is not None:
        val = shape[axis] - end_slice.start
        window = func(2 * val)[val:]
        data_inds = broadcast_for_index(n_dims, axis, end_slice)
        window_inds = broadcast_for_index(n_dims, axis, slice(None), fill=None)
        out[data_inds] = out[data_inds] * window[window_inds]
    return patch.new(data=out)



"""Module for performing integrations."""
from __future__ import annotations

import numpy as np
import pytest

import dascore as dc
import dascore.proc.coords
from dascore.transform.integrate import integrate
from dascore.units import get_quantity
from dascore.utils.misc import broadcast_for_index
from dascore.utils.time import to_float


@pytest.fixture(scope="session")
def ones_patch(random_patch):
    """Return a patch of ones with nornal axis."""
    array = np.ones_like(random_patch.data)
    return random_patch.new(data=array)


class TestIndefiniteIntegrals:
    """Tests for indefinite integrals."""

    def test_indef_integration(self, ones_patch):
        """Happy path for default time/distance integrals."""
        for dim in ones_patch.dims:
            patch = integrate(ones_patch, dim=dim, definite=False)
            ax = patch.dims.index(dim)
            # We expect slice not on axis to be the identical
            non_dim_indexer = broadcast_for_index(
                len(patch.dims), ax, slice(None, None), fill=slice(0, 1)
            )
            first_along_axis = patch.data[non_dim_indexer]
            assert np.allclose(patch.data, first_along_axis)
            # the values along the axis of integration should strictly increase.
            # (because they are all 1s)
            flat = np.squeeze(first_along_axis)
            assert np.allclose(flat, np.sort(flat))

    def test_integrate_non_evenly_sampled(self, wacky_dim_patch):
        """Ensure we can integrate along non-evenly sampled dims."""
        out = wacky_dim_patch.integrate(dim="time", definite=False)
        assert isinstance(out, dc.Patch)

    def test_units(self, random_patch):
        """Ensure output units are as expected."""
        patch = random_patch.set_units("m/s")
        dims = ("time", "distance")
        expected_units = ("s", "m")
        for dim, eu in zip(dims, expected_units):
            out = patch.integrate(dim=dim)
            data_units1 = get_quantity(patch.attrs.data_units)
            data_units2 = get_quantity(out.attrs.data_units)
            assert data_units2 == (data_units1 * get_quantity(eu))
            for dim in patch.dims:
                coord1 = patch.get_coord(dim)
                coord2 = patch.get_coord(dim)
                assert coord2.units == coord1.units


class TestDefiniteIntegration:
    """Test case for definite path integration."""

    def test_simple_integration(self, ones_patch):
        """Ensure simple integration works."""
        patch = ones_patch
        for dim in patch.dims:
            ax = patch.dims.index(dim)
            out = patch.integrate(dim=dim, definite=True)
            assert out.shape[ax] == 1
            step = to_float(patch.get_coord(dim).step)
            expected_data = np.trapz(patch.data, dx=step, axis=ax)
            ndims = len(patch.dims)
            indexer = broadcast_for_index(ndims, ax, None)
            assert np.allclose(out.data, expected_data[indexer])
            # Since the patch is just ones all values should equal the
            # dimensional length when dx == 1
            if dim == "distance" and np.isclose(patch.get_coord(dim).step, 1):
                assert np.allclose(out.data, patch.shape[ax] - 1)

    def test_units(self, random_patch):
        """Ensure data units are updated and coord units are unchanged."""
        patch = random_patch.set_units("m/s")
        out = patch.integrate(dim="time", definite=True)
        data_units1 = get_quantity(patch.attrs.data_units)
        data_units2 = get_quantity(out.attrs.data_units)
        assert data_units2 == (data_units1 * get_quantity("s"))
        for dim in patch.dims:
            coord1 = patch.get_coord(dim)
            coord2 = patch.get_coord(dim)
            assert coord2.units == coord1.units

    def test_integrate_all_dims(self, random_patch):
        """Ensure all dims can be integrated."""
        out = random_patch.integrate(dim=None, definite=True)
        assert out.shape == tuple([1] * len(random_patch.shape))

    def test_integrate_non_evenly_sampled_dim(self, wacky_dim_patch):
        """Simple test to integrate along non-evenly sampled dimension."""
        out = wacky_dim_patch.integrate(dim="time", definite=True)
        assert isinstance(out, dc.Patch)



"""Misc Utilities."""
from __future__ import annotations

import contextlib
import functools
import importlib
import inspect
import os
import re
import warnings
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from functools import cache
from pathlib import Path
from types import ModuleType

import numpy as np
import pandas as pd
from scipy.linalg import solve
from scipy.special import factorial

import dascore as dc
from dascore.exceptions import (
    FilterValueError,
    MissingOptionalDependency,
    ParameterError,
)
from dascore.utils.progress import track


class _Sentinel:
    """Sentinel for key checks."""


def register_func(list_or_dict: list | dict, key=None):
    """
    Decorator for registering a function name in a list or dict.

    If list_or_dict is a list only append the name of the function. If it is
    as dict append name (as key) and function as the value.

    Parameters
    ----------
    list_or_dict
        A list or dict to which the wrapped function will be added.
    key
        The name to use, if different than the name of the function.
    """

    def wrapper(func):
        name = key or func.__name__
        if hasattr(list_or_dict, "append"):
            list_or_dict.append(name)
        else:
            list_or_dict[name] = func
        return func

    return wrapper


def _pass_through_method(func):
    """Decorator for marking functions as methods on namedspace parent class."""

    @functools.wraps(func)
    def _func(self, *args, **kwargs):
        obj = self._obj
        return func(obj, *args, **kwargs)

    return _func


class _NameSpaceMeta(type):
    """Metaclass for namespace class."""

    def __setattr__(self, key, value):
        if callable(value):
            value = _pass_through_method(value)
        super().__setattr__(key, value)


@contextlib.contextmanager
def suppress_warnings(category=Warning):
    """
    Context manager for suppressing warnings.

    Parameters
    ----------
    category
        The types of warnings to suppress. Must be a subclass of Warning.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=category)
        yield
    return None


class MethodNameSpace(metaclass=_NameSpaceMeta):
    """A namespace for class methods."""

    def __init__(self, obj):
        self._obj = obj

    def __init_subclass__(cls, **kwargs):
        """Wrap all public methods."""
        for key, val in vars(cls).items():
            if callable(val):  # passes to _NameSpaceMeta settattr
                setattr(cls, key, val)


def broadcast_for_index(
    n_dims: int,
    axis: int | Sequence[int],
    value: slice | int | None,
    fill=slice(None),
):
    """
    For a given shape of array, return empty slices except for slice axis.

    Parameters
    ----------
    n_dims
        The number of dimensions in the array that will be indexed.
    axis
        The axis number.
    value
        A slice object.
    fill
        The default values for non-axis entries.
    """
    axes = set(iterate(axis))
    return tuple(fill if x not in axes else value for x in range(n_dims))


def all_close(ar1, ar2):
    """
    Return True if ar1 is allcose to ar2.

    Just uses numpy.allclose unless ar1 is a datetime, in which case
    strict equality is used.
    """
    ar1, ar2 = np.array(ar1), np.array(ar2)
    is_date = np.issubdtype(ar1.dtype, np.datetime64)
    is_timedelta = np.issubdtype(ar1.dtype, np.timedelta64)
    if is_date or is_timedelta:
        return np.all(ar1 == ar2)
    else:
        return np.allclose(ar1, ar2)


def iter_files(
    paths: str | Iterable[str],
    ext: str | None = None,
    mtime: float | None = None,
    skip_hidden: bool = True,
) -> Iterable[str]:
    """
    Use os.scan dir to iter files, optionally only for those with given
    extension (ext) or modified times after mtime.

    Parameters
    ----------
    paths
        The path to the base directory to traverse. Can also use a collection
        of paths.
    ext : str or None
        The extensions to map.
    mtime : int or float
        Time stamp indicating the minimum mtime.
    skip_hidden : bool
        If True skip files or folders (they begin with a '.')

    Yields
    ------
    Paths, as strings, meeting requirements.
    """
    try:  # a single path was passed
        for entry in os.scandir(paths):
            if entry.is_file() and (ext is None or entry.name.endswith(ext)):
                if mtime is None or entry.stat().st_mtime >= mtime:
                    if entry.name[0] != "." or not skip_hidden:
                        yield entry.path
            elif entry.is_dir() and not (skip_hidden and entry.name[0] == "."):
                yield from iter_files(
                    entry.path, ext=ext, mtime=mtime, skip_hidden=skip_hidden
                )
    except TypeError:  # multiple paths were passed
        for path in paths:
            yield from iter_files(path, ext, mtime, skip_hidden)
    except NotADirectoryError:  # a file path was passed, just return it
        yield paths


def iterate(obj):
    """
    Return an iterable from any object.

    If a string, do not iterate characters, return str in tuple.

    *This is how iteration *should* work in python.
    """
    if obj is None:
        return ()
    if isinstance(obj, str):
        return (obj,)
    return obj if isinstance(obj, Iterable) else (obj,)


class CacheDescriptor:
    """A descriptor for storing infor in an instance cache (mapping)."""

    def __init__(self, cache_name, func_name, args=None, kwargs=None):
        self._cache_name = cache_name
        self._func_name = func_name
        self._args = () if args is None else args
        self._kwargs = {} if kwargs is None else kwargs

    def __set_name__(self, owner, name):
        """Method to set the name of the description on the instance."""
        self._name = name

    def __get__(self, instance, owner):
        """Get contents of the cache."""
        cache = getattr(instance, self._cache_name)
        if self._name not in cache:
            func = getattr(instance, self._func_name)
            out = func(*self._args, **self._kwargs)
            cache[self._name] = out
        return cache[self._name]

    def __set__(self, instance, value):
        """Set the cache contents."""
        cache = getattr(instance, self._cache_name)
        cache[self._name] = value


def optional_import(package_name: str) -> ModuleType:
    """
    Import a module and return the module object if installed, else raise error.

    Parameters
    ----------
    package_name
        The name of the package which may or may not be installed. Can
        also be sub-packages/modules (eg dascore.core).

    Raises
    ------
    MissingOptionalDependency if the package is not installed.

    Examples
    --------
    >>> from dascore.utils.misc import optional_import
    >>> from dascore.exceptions import MissingOptionalDependency
    >>> # import a module (this is the same as import dascore as dc)
    >>> dc = optional_import('dascore')
    >>> try:
    ...     optional_import('boblib5')  # doesn't exist so this raises
    ... except MissingOptionalDependency:
    ...     pass
    """
    try:
        mod = importlib.import_module(package_name)
    except ImportError:
        msg = (
            f"{package_name} is not installed but is required for the "
            f"requested functionality"
        )
        raise MissingOptionalDependency(msg)
    return mod


def get_middle_value(array):
    """Get the middle value in the differences array without changing dtype."""
    array = np.sort(np.array(array))
    last_ind = len(array) - 1
    ind = int(np.floor(last_ind / 2))
    return np.sort(array)[ind]


def all_diffs_close_enough(diffs):
    """Check if all the diffs are 'close' handling timedeltas."""
    if not len(diffs):
        return False
    diffs = np.array(diffs)
    is_dt = np.issubdtype(diffs.dtype, np.timedelta64)
    is_td = np.issubdtype(diffs.dtype, np.datetime64)
    if is_td or is_dt:
        diffs = diffs.astype(np.int64).astype(np.float64)
    med = np.median(diffs)
    # Note: The rtol parameter here is a bit arbitrary; it was set
    # based on experience but there is probably a better way to do this.
    return np.allclose(diffs, med, rtol=0.001)


def unbyte(byte_or_str: bytes | str) -> str:
    """Ensure a string is given by str or possibly bytes."""
    if isinstance(byte_or_str, bytes | np.bytes_):
        byte_or_str = byte_or_str.decode("utf8")
    return byte_or_str


def _get_stencil_weights(array, ref_point, order):
    """
    Computes the derivative stencil weights.

    Parameters
    ----------
        array
            An array representing the stencil domain.
        ref_point
            The point in the domain to base the stencil weights on.
        order
            The order of the derivative.

    Returns
    -------
        The vector of stencil weights.
    """
    ell = np.arange(len(array))
    assert order in ell, "Order must be in domain"
    A = (((array - ref_point)[:, np.newaxis] ** ell) / factorial(ell)).T
    weights = solve(A, ell == order)
    return weights.flatten()


def _maybe_make_parent_directory(path):
    """Maybe make parent directories."""
    path = Path(path)
    if path.exists():
        return
    path.parent.mkdir(parents=True, exist_ok=True)
    return


def get_stencil_coefs(order, derivative=2):
    """Get centered coefficients for a derivative of specified order and derivative."""
    dx = np.arange(-order, order + 1)
    return _get_stencil_weights(dx, 0, derivative)


def get_parent_code_name(levels: int = 2) -> str:
    """Get the name of the calling function/class levels up in stack."""
    stack = inspect.currentframe()
    for _ in range(levels):
        stack = stack.f_back
    return stack.f_code.co_name


def to_str(val):
    """Convert value to string."""
    # This is primarily used to avoid lambdas which can cause issues
    # in pickling.
    return str(val)


def maybe_get_attrs(obj, attr_map: Mapping):
    """Maybe get attributes from object (if they exist)."""
    out = {}
    for old_name, new_name in attr_map.items():
        if hasattr(obj, old_name):
            value = getattr(obj, old_name)
            out[new_name] = unbyte(value)
    return out


def maybe_get_items(obj, attr_map: Mapping):
    """Maybe get items from a mapping (if they exist)."""
    out = {}
    for old_name, new_name in attr_map.items():
        if not (value := obj.get(old_name, None)):
            continue
        out[new_name] = unbyte(value)
    return out


@cache
def _get_compiled_suffix_prefix_regex(
    suffixes: str | tuple[str],
    prefixes: str | tuple[str] | None,
):
    """Get a compiled regex which matches the form prefixes_suffixes."""
    suffixes = iterate(suffixes)
    pattern = rf".*_({'|'.join(iterate(suffixes))})"
    if prefixes is not None:
        pattern = rf"({'|'.join(iterate(prefixes))})" + pattern
    return re.compile(pattern)


def _matches_prefix_suffix(input_str, suffixes, prefixes=None):
    """Determine if a string matches given prefixes_suffixes."""
    regex = _get_compiled_suffix_prefix_regex(suffixes, prefixes)
    return bool(re.match(regex, input_str))


def is_valid_coord_str(input_str, prefixes=None):
    """Return True if an input string is valid for representing coord info."""
    _valid_keys = tuple(dc.core.CoordSummary.model_fields)
    return _matches_prefix_suffix(input_str, _valid_keys, prefixes)


def separate_coord_info(
    obj,
    dims: tuple[str] | None = None,
    required: Sequence[str] | None = None,
    cant_be_alone: tuple[str] = ("units", "dtype"),
) -> tuple[dict, dict]:
    """
    Separate coordinate information from attr dict.

    These can be in the flat-form (ie {time_min, time_max, time_step, ...})
    or a nested coord: {coords: {time: {min, max, step}}

    Parameters
    ----------
    obj
        The object or model to
    dims
        The dimension to look for.
    required
        If provided, the required attributes (e.g., min, max, step).
    cant_be_alone
        names which cannot be on their own.

    Returns
    -------
    coord_dict and attrs_dict.
    """

    def _meets_required(coord_dict):
        """Return True coord dict meets the minimum required keys."""
        if not coord_dict:
            return False
        if not required and (set(coord_dict) - cant_be_alone):
            return True
        return set(coord_dict).issuperset(required)

    def _get_dims(obj):
        """Try to ascertain dims from keys in obj."""
        potential_keys = defaultdict(set)
        for key in obj:
            if not is_valid_coord_str(key):
                continue
            potential_keys[key.split("_")[0]].add(key.split("_")[1])
        return tuple(i for i, v in potential_keys.items() if _meets_required(v))

    def _get_coords_from_top_level(obj, out, dims):
        """First get coord info from top level."""
        for dim in iterate(dims):
            potential_coord = {
                i.split("_")[1]: v for i, v in obj.items() if is_valid_coord_str(i, dim)
            }
            # nasty hack for handling d_{dim} for backward compatibility.
            if (bad_name := f"d_{dim}") in obj:
                msg = f"d_{dim} is deprecated, use {dim}_step"
                warnings.warn(msg, DeprecationWarning, stacklevel=3)
                potential_coord["step"] = obj[bad_name]

            if _meets_required(potential_coord):
                out[dim] = potential_coord

    def _get_coords_from_coord_level(obj, out):
        """Get coords from coordinate level."""
        coords = obj.get("coords", {})
        if hasattr(coords, "to_summary_dict"):
            coords = coords.to_summary_dict()
        for key, value in coords.items():
            if hasattr(value, "to_summary"):
                value = value.to_summary()
            if hasattr(value, "model_dump"):
                value = value.model_dump()
            if _meets_required(value):
                out[key] = value

    def _pop_keys(obj, out):
        """Pop out old keys for attrs, and unused keys from out."""
        # first coord subdict
        obj.pop("coords", None)
        # then top-level
        for coord_name, sub_dict in out.items():
            for thing_name in sub_dict:
                obj.pop(f"{coord_name}_{thing_name}", None)
            if "step" in sub_dict:
                obj.pop(f"d_{coord_name}", None)

    # sequence of short-circuit checks
    out = {}
    required = set(required) if required is not None else set()
    cant_be_alone = set(cant_be_alone)
    if obj is None:
        return out, {}
    if hasattr(obj, "model_dump"):
        obj = obj.model_dump()
    if dims is None:
        dims = _get_dims(obj)
    # this is already a dict of coord info.
    if set(dims) == set(obj):
        return obj, {}
    obj = dict(obj)
    _get_coords_from_coord_level(obj, out)
    _get_coords_from_top_level(obj, out, dims)
    _pop_keys(obj, out)
    return out, obj


def cached_method(func):
    """
    Cache decorated method.

    Simply uses the id of self for the key rather than hashing it.
    We can't use functools.cache due to pydantic #6787.
    """
    sentinel = object()  # unique object for cache misses.

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not hasattr(self, "_cache"):
            self._cache = {}
        cache = self._cache
        if not (args or kwargs):
            key = id(func)
        else:
            key = (id(func), *args)
            if kwargs:
                for item in kwargs.items():
                    key += item
        out = cache.get(key, sentinel)
        if out is not sentinel:
            return out
        out = func(self, *args, **kwargs)
        cache[key] = out
        return out

    return wrapper


class _MapFuncWrapper:
    """A class for unwrapping spools to base applies."""

    def __init__(self, func, kwargs, progress=True):
        self._func = func
        self._kwargs = kwargs
        self._progress = progress

    def __call__(self, spool):
        iterable = spool
        # in order to handle multiprocessing, we apply a secret tag of "_progress"
        # to the first spool. This way only the first spool displays the
        # the progress bar. A huge hack, maybe there is a better way? See #265.
        if not getattr(spool, "_no_progress", False):
            desc = f"Applying {self._func.__name__} to spool"
            iterable = track(spool, desc) if self._progress else spool
        return [self._func(x, **self._kwargs) for x in iterable]


def _spool_map(spool, func, size=None, client=None, progress=True, **kwargs):
    """
    Map a func over a spool.

    Parameters
    ----------
    spool
        The spool object ot apply func to
    size
        The number of patches for each spool (ie chunksize)
    client
        An object with a map method for applying concurrency.
    progress
        If True, display a progress bar.
    **kwargs
        Keywords passed to func.
    """
    # no client; simple for loop.
    desc = f"Applying {func.__name__} to spool"
    if client is None:
        iterable = track(spool, desc) if progress else spool
        return [func(patch, **kwargs) for patch in iterable]
    # Now things get interesting. We need to split the spool here
    # so that patches don't get serialized.
    if size is None:
        size = len(spool) / os.cpu_count()
    spools = list(spool.split(size=size))
    # this is a hack to get the progress bar to work. Essentially, we just
    # add a secret flag to all but one spool so that progress bar is only
    # displayed in one thread/process.
    for sub_spool in spools[1:]:
        sub_spool._no_progress = True
    new_func = _MapFuncWrapper(func, kwargs, progress=progress)
    return [x for y in client.map(new_func, spools) for x in y]


def _dict_list_diffs(dict_list):
    """Return the keys which are not equal dicts in a list."""
    out = set()
    first = dict_list[0]
    first_keys = set(first)
    for other in dict_list[1:]:
        if other == first:
            continue
        other_keys = set(other)
        out |= (other_keys - first_keys) | (first_keys - other_keys)
        common_keys = other_keys & first_keys
        for key in common_keys:
            if first[key] != other[key]:
                out.add(key)
    return sorted(out)


def sanitize_range_param(select) -> tuple:
    """Given a slice or tuple, check and return slice or tuple."""
    # handle slices, need to convert to tuple
    if isinstance(select, slice):
        if select.step is not None:
            msg = (
                "Step not supported in select/filtering. Use decimate for "
                "proper down-sampling."
            )
            raise ParameterError(msg)
        select = (select.start, select.stop)
    # convert ellipses or ellipses values
    if select is None or select is Ellipsis:
        select = (None, None)
    # validate length (only length 2 allowed)
    if len(select) != 2:
        msg = "Range indices must be a length 2 sequence."
        raise ParameterError(msg)
    # swap out ellipses for None so downstream funcs dont have to
    select = tuple(None if x is ... else x for x in select)
    return select


def check_filter_kwargs(kwargs):
    """Check filter kwargs and return dim name and filter range."""
    if len(kwargs) != 1:
        msg = "pass filter requires you specify one dimension and filter range."
        raise FilterValueError(msg)
    dim = next(iter(kwargs.keys()))
    filt_range = kwargs[dim]
    # strip out units if used.
    mags = tuple([getattr(x, "magnitude", x) for x in filt_range])
    if not isinstance(filt_range, Sequence) or len(filt_range) != 2:
        msg = f"filter range must be a length two sequence not {filt_range}"
        raise FilterValueError(msg)
    if all([pd.isnull(x) for x in mags]):
        msg = (
            f"pass filter requires at least one filter limit, "
            f"you passed {filt_range}"
        )
        raise FilterValueError(msg)

    return dim, filt_range


def check_filter_range(nyquist, low, high, filt_min, filt_max):
    """Simple check on filter parameters."""
    # ensure filter bounds are within nyquist
    if low is not None and ((0 > low) or (low > 1)):
        msg = f"possible filter bounds are [0, {nyquist}] you passed {filt_min}"
        raise FilterValueError(msg)
    if high is not None and ((0 > high) or (high > 1)):
        msg = f"possible filter bounds are [0, {nyquist}] you passed {filt_max}"
        raise FilterValueError(msg)
    if high is not None and low is not None and high <= low:
        msg = (
            "Low filter param must be less than high filter param, you passed:"
            f"filt_min = {filt_min}, filt_max = {filt_max}"
        )
        raise FilterValueError(msg)



"""Generic tests for prodml support."""
from __future__ import annotations

import shutil

import pytest
import tables

import dascore as dc
from dascore.io.core import read
from dascore.utils.downloader import fetch


@pytest.fixture(scope="session")
def quantx_v2_example_path():
    """Return the path to the example QuantXV2 file."""
    out = fetch("opta_sense_quantx_v2.h5")
    assert out.exists()
    return out


@pytest.fixture(scope="session")
def quantx_v2_das_patch(quantx_v2_example_path):
    """Read the QuantXV2 data, return contained DataArray."""
    out = read(quantx_v2_example_path, "prodml")[0]
    attr_time = out.attrs["time_max"]
    coord_time = out.coords.max("time")
    assert attr_time == coord_time
    return out


class TestProdMLFile:
    """
    Ensure we can read the ProdML provided by Silixa.

    We do this since the other tests read the prodML files, even though
    the Silixa file is technical just ProdML v2.1.
    """

    @pytest.fixture(scope="class")
    def issue_221_patch_path(self, tmp_path_factory):
        """Ensure dims are correctly ascertained."""
        tmp_path = tmp_path_factory.mktemp("issue_221")
        path = dc.utils.downloader.fetch("prodml_2.0.h5")
        new_path = shutil.copy2(path, tmp_path / "prod_2_monkey_patched.h5")
        with tables.open_file(new_path, "a") as fi:
            # monkey patch dimensions to simulate issue.
            new_dims = "time, locus"
            parent_node = fi.root.Acquisition["Raw[0]"]
            node = parent_node["RawData"]
            node._v_attrs.Dimensions = new_dims
        return new_path

    @pytest.fixture(scope="class")
    def silixa_h5_patch(self, idas_h5_example_path):
        """Get the silixa file, return Patch."""
        return dc.spool(idas_h5_example_path)[0]

    def test_read_silixa(self, silixa_h5_patch):
        """Ensure we can read  Silixa file."""
        assert isinstance(silixa_h5_patch, dc.Patch)
        assert silixa_h5_patch.shape

    def test_has_gauge_length(self, silixa_h5_patch):
        """Ensure gauge-length is found in patch attrs."""
        patch = silixa_h5_patch
        assert hasattr(patch.attrs, "gauge_length")

    def test_issue_221(self, issue_221_patch_path):
        """Ensure dims are correctly ascertained."""
        patch = dc.read(issue_221_patch_path)[0]
        assert isinstance(patch, dc.Patch)


class TestReadQuantXV2:
    """Tests for reading the QuantXV2 format."""

    def test_precision_of_time_array(self, quantx_v2_das_patch):
        """
        Ensure the time array is in ns, not native ms, in order to
        be consistent with other patches.
        """
        time = quantx_v2_das_patch.coords.get_array("time")
        dtype = time.dtype
        assert "[ns]" in str(dtype)



"""Tests for displaying dascore objects."""
from __future__ import annotations

import numpy as np
import pandas as pd

import dascore as dc
from dascore.utils.display import get_nice_text


class TestGetNiceText:
    """Tests for converting coordinate to nice looking rich Text."""

    def test_simple_datetime(self):
        """Ensure the process works for datetime objects."""
        dt = dc.to_datetime64("2023-10-01")
        # YMD should just show YMD
        txt1 = get_nice_text(dt)
        assert str(txt1) == "2023-10-01"
        # Unless YMD is 1970-01-01
        txt2 = get_nice_text(dc.to_datetime64(0))
        assert str(txt2) == "00:00:00"
        # Decimals are displayed if present
        txt3 = get_nice_text(dc.to_datetime64(1.111111111))
        assert str(txt3).endswith(".111111111")

    def test_nat(self):
        """Tests for NaT."""
        dt = np.datetime64("NaT")
        txt = get_nice_text(dt)
        assert str(txt) == "NaT"

    def test_timestamp(self):
        """Tests for pandas timestamps."""
        ts = pd.Timestamp("2012-01-10")
        txt = get_nice_text(ts)
        assert str(txt) == "2012-01-10"



"""Core module for reading and writing pickle format."""
from __future__ import annotations

import pickle

import dascore
from dascore.io import BinaryReader, BinaryWriter, FiberIO


class PickleIO(FiberIO):
    """
    Provides IO support for the pickle format.

    Warning
    -------
    The pickle format is discouraged due to potential security and
    compatibility issues.
    """

    name = "PICKLE"
    preferred_extensions = ("pkl", "pickle")

    def _header_is_dascore(self, byte_stream):
        """Return True if the first few bytes mention dascore classes."""
        has_dascore = b"dascore.core" in byte_stream
        spool_or_patch = b"Spool" in byte_stream or b"Patch" in byte_stream
        return has_dascore and spool_or_patch

    def get_format(self, resource: BinaryReader) -> tuple[str, str] | bool:
        """
        Return True if file contains a pickled Patch or Spool.

        Parameters
        ----------
        resource
            A path to the file which may contain terra15 data.
        """
        try:
            start = resource.read(100)  # read first 100 bytes, look for class names
            if self._header_is_dascore(start):
                getattr(resource, "seek", lambda x: None)(0)
                pickle.load(resource)
                return ("PICKLE", self.version)  # TODO add pickle protocol
            else:
                return False
        except (pickle.UnpicklingError, FileNotFoundError, IndexError):
            return False

    def read(self, resource: BinaryReader, **kwargs):
        """Read a Patch/Spool from disk."""
        out = pickle.load(resource)
        return dascore.spool(out)

    def write(self, patch, resource: BinaryWriter, **kwargs):
        """Write a Patch/Spool to disk."""
        pickle.dump(patch, resource)



"""Test for basic IO and related functions."""
from __future__ import annotations

import copy
import io
from pathlib import Path
from typing import TypeVar

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.constants import SpoolType
from dascore.exceptions import InvalidFiberIO, UnknownFiberFormat
from dascore.io.core import FiberIO, PatchFileSummary
from dascore.io.dasdae.core import DASDAEV1
from dascore.utils.io import BinaryReader, BinaryWriter
from dascore.utils.time import to_datetime64

tvar = TypeVar("tvar", int, float, str, Path)


class _FiberFormatTestV1(FiberIO):
    """A test format v1."""

    name = "_TestFormatter"
    version = "1"


class _FiberFormatTestV2(FiberIO):
    """A test format v2."""

    name = "_TestFormatter"
    version = "2"


class _FiberImplementer(FiberIO):
    """A fiber io which implements all the methods (poorly)."""

    name = "_Implementer"
    version = "2"

    def read(self, resource, **kwargs):
        """Dummy read."""

    def write(self, spool: SpoolType, resource):
        """Dummy write."""

    def scan(self, resource: BinaryReader):
        """Dummy scan."""

    def get_format(self, resource):
        """Dummy get_format."""


class _FiberCaster(FiberIO):
    """A test class for casting inputs to certain types."""

    name = "_TestFormatter"
    version = "2"

    def read(self, resource: BinaryReader, **kwargs) -> SpoolType:
        """Just ensure read was cast to correct type."""
        assert isinstance(resource, io.BufferedReader)

    def write(self, spool: SpoolType, resource: BinaryWriter):
        """Ditto for write."""
        assert isinstance(resource, io.BufferedWriter)

    def get_format(self, resource: Path) -> tuple[str, str] | bool:
        """And get format."""
        assert isinstance(resource, Path)
        return False

    def scan(self, not_path: BinaryReader):
        """Ensure an off-name still works for type casting."""
        assert isinstance(not_path, io.BufferedReader)


class _FiberUnsupportedTypeHints(FiberIO):
    """A fiber io which implements typehints which have not casting meaning."""

    name = "_TypeHinterNotRight"
    version = "2"

    def read(self, resource: tvar, **kwargs):
        """Dummy read."""
        with open(resource) as fi:
            return fi.read()


class TestPatchFileSummary:
    """Tests for getting patch file information."""

    def test_d_translates(self):
        """Ensure d_{whatever} translates to step_{whatever}."""
        out = PatchFileSummary(d_time=10)
        assert out.time_step == dc.to_timedelta64(10)

    def test_dim_typle(self):
        """Ensure patch file summaries can be converted to tuples."""
        out = PatchFileSummary(d_time=10, dims="time,distance")
        assert out.dim_tuple == ("time", "distance")

    def test_flat_dump(self):
        """Simple test to show summary can be flat dumped."""
        # flat dump is just here for compatibility with dc.PatchAttrs
        out = PatchFileSummary(d_time=10, dims="time,distance")
        assert isinstance(out.flat_dump(), dict)


class TestFormatManager:
    """Tests for the format manager."""

    @pytest.fixture(scope="class")
    def format_manager(self):
        """Deep copy manager to avoid changing state used by other objects."""
        manager = copy.deepcopy(FiberIO.manager)
        return manager

    def test_specific_format_and_version(self, format_manager):
        """Specifying a known format and version should return exactly one formatter."""
        out = list(format_manager.yield_fiberio("DASDAE", "1"))
        assert len(out) == 1
        assert isinstance(out[0], DASDAEV1)

    def test_get_all_formatters(self, format_manager):
        """Ensure getting all formatters through yield_fiberio works."""
        formatters = list(format_manager.yield_fiberio())
        assert len(formatters) >= len(format_manager._eps)

    def test_extension_priority(self, format_manager):
        """Ensure the extension priority is honored."""
        ext = "h5"
        ext_formatters = list(format_manager.yield_fiberio(extension=ext))
        all_formatters = list(format_manager.yield_fiberio())
        in_formatter = [ext in x.preferred_extensions for x in ext_formatters]
        format_array = np.array(in_formatter).astype(bool)
        # ensure all the start of the arrays are True.
        assert np.argmin(format_array) == np.sum(format_array)
        # ensure all formats are represented.
        assert len(format_array) == len(all_formatters)
        # ensure V2 of the Test formatter appears first
        v2_arg = np.argmax([isinstance(x, _FiberImplementer) for x in ext_formatters])
        v1_arg = np.argmax([isinstance(x, _FiberFormatTestV1) for x in ext_formatters])
        assert v2_arg < v1_arg

    def test_format_raises_unknown_format(self, format_manager):
        """Ensure we raise for unknown formats."""
        with pytest.raises(UnknownFiberFormat, match="format"):
            list(format_manager.yield_fiberio(format="bob_2"))

    def test_format_raises_just_version(self, format_manager):
        """Providing only a version should also raise."""
        with pytest.raises(UnknownFiberFormat, match="version"):
            list(format_manager.yield_fiberio(version="1"))

    def test_format_bad_version(self, format_manager):
        """Ensure providing a bad version but valid format raises."""
        with pytest.raises(UnknownFiberFormat, match="known versions"):
            iterator = format_manager.yield_fiberio(format="DASDAE", version="-1")
            list(iterator)

    def test_format_format_no_version(self, format_manager):
        """Ensure providing a bad version but valid format raises."""
        with pytest.raises(UnknownFiberFormat, match="known versions"):
            iterator = format_manager.yield_fiberio(format="DASDAE", version="-1")
            list(iterator)

    def test_format_multiple_versions(self, format_manager):
        """Ensure multiple versions are returned when only format is specified."""
        file_format = _FiberFormatTestV1.name
        out = list(format_manager.yield_fiberio(format=file_format))
        assert len(out) == 2


class TestFormatter:
    """Tests for adding file supports through Formatter."""

    # the methods a formatter can implement.

    class FormatterWithName(FiberIO):
        """A formatter with a file name."""

        name = "_test_format"

    def test_empty_formatter_raises(self):
        """An empty formatter can't exist; it at least needs a name."""
        with pytest.raises(InvalidFiberIO):

            class empty_formatter(FiberIO):
                """formatter with no name."""

    def test_empty_formatter_undefined_methods(self, random_patch):
        """
        Ensure a Not Implemented error is raised for un-implemented methods
        of FormatterWithName.
        """
        instance = self.FormatterWithName()
        with pytest.raises(NotImplementedError):
            instance.read("empty_path")
        with pytest.raises(NotImplementedError):
            instance.write(random_patch, "empty_path")
        with pytest.raises(NotImplementedError):
            instance.get_format("empty_path")
        with pytest.raises(NotImplementedError):
            instance.scan("bad_path")

    def test_doesnt_implements(self):
        """Tests for implements_x methods."""
        # this test fiber io don't implement anything
        fio = _FiberFormatTestV1()
        assert not fio.implements_scan
        assert not fio.implements_get_format
        assert not fio.implements_read
        assert not fio.implements_write

    def test_implements(self):
        """Tests for implements_x methods."""
        # this test fiber implements all the things
        fio = _FiberImplementer()
        assert fio.implements_scan
        assert fio.implements_get_format
        assert fio.implements_read
        assert fio.implements_write


class TestGetFormat:
    """Tests to ensure formats can be retrieved."""

    def test_not_known(self, dummy_text_file):
        """Ensure a non-path/str object raises."""
        with pytest.raises(UnknownFiberFormat):
            dc.get_format(dummy_text_file)

    def test_missing_file(self):
        """Ensure a missing file raises."""
        with pytest.raises(FileNotFoundError):
            dc.get_format("bad/file")


class TestScan:
    """Tests for scanning fiber files."""

    @pytest.fixture(scope="class")
    def nested_directory_with_patches(self, tmpdir_factory, random_patch):
        """Return a nested directory with patch files interlaced."""
        out = Path(tmpdir_factory.mktemp("nested_random_patch"))
        path_1 = out / "patch_1.h5"
        path_2 = out / "subdir" / "patch_2.h5"
        path_3 = out / "subdir" / "suber_dir" / "patch_3.h5"
        random_patch.io.write(path_1, "dasdae")
        random_patch.io.write(path_2, "dasdae")
        random_patch.io.write(path_3, "dasdae")
        return out

    def test_scan_no_good_files(self, tmp_path):
        """Scan with no fiber files should return []."""
        dummy_file = tmp_path / "data.txt"
        dummy_file.touch()
        out = dc.scan(dummy_file)
        assert not len(out)
        assert out == []

    def test_scan_directory(self, tmp_path):
        """Trying to scan an empty directory should return empty list."""
        out = dc.scan(tmp_path)
        assert len(out) == 0

    def test_scan_bad_files(self, tmp_path):
        """Trying to scan a directory should raise a nice error."""
        new = tmp_path / "myfile.txt"
        with pytest.raises(FileNotFoundError):
            _ = dc.scan(new)

    def test_scan_patch(self, random_patch):
        """Scan should also work on a patch."""
        out = dc.scan_to_df(random_patch)
        attrs = random_patch.attrs
        assert len(out) == 1
        ser = out.iloc[0]
        assert to_datetime64(ser["time_min"]) == to_datetime64(attrs["time_min"])
        assert to_datetime64(ser["time_max"]) == to_datetime64(attrs["time_max"])

    def test_scan_nested_directory(self, nested_directory_with_patches):
        """Ensure scan picks up files in nested directories."""
        out = dc.scan(nested_directory_with_patches)
        assert len(out) == 3

    def test_can_raise(self):
        """
        Scan, when called from a FiberIO, should be able to raise if
        type coercion fails.
        """
        fio = _FiberImplementer()
        bad_input = _FiberFormatTestV1()
        with pytest.raises(NotImplementedError):
            fio.scan(bad_input)


class TestCastType:
    """Test suite to ensure types are intelligently cast to type hints."""

    def test_read(self, dummy_text_file):
        """Ensure write casts type."""
        io = _FiberCaster()
        # this passes if it doesnt raise.
        io.read(dummy_text_file)

    def test_write(self, tmp_path, random_spool):
        """Ensure write casts type."""
        path = tmp_path / "write_fiber_cast.txt"
        io = _FiberCaster()
        # this passes if it doesnt raise.
        io.write(random_spool, path)

    def test_non_standard_name(self, dummy_text_file):
        """Ensure non-standard names still work."""
        io = _FiberCaster()
        io.scan(dummy_text_file)

    def test_unsupported_typehints(self, dummy_text_file):
        """Ensure FiberIO with non-"special" type hints still works."""
        fiberio = _FiberUnsupportedTypeHints()
        out = fiberio.read(dummy_text_file)
        assert out == Path(dummy_text_file).read_text()

    def test_unsupported_type(self, dummy_text_file):
        """Ensure FiberIO from above works with dascore.read."""
        name = _FiberUnsupportedTypeHints.name
        version = _FiberUnsupportedTypeHints.version
        out = dc.read(dummy_text_file, name, version)
        assert out == Path(dummy_text_file).read_text()


class TestGetSupportedIOTable:
    """A test for creating the supported io table."""

    def test_get_supported_io_table(self):
        """Test the get_supported_io_table function."""
        # call the function to get the result
        result_df = FiberIO.get_supported_io_table()

        # assert that the result is a DataFrame
        assert isinstance(result_df, pd.DataFrame)

        # assert that the length of the DataFrame is not 0
        assert len(result_df) > 0



"""Module for applying aggregations along a specified axis."""
from __future__ import annotations

from functools import partial
from typing import Literal

import numpy as np

from dascore.constants import PatchType
from dascore.utils.patch import patch_function
from dascore.utils.time import to_datetime64


def _take_first(data, axis):
    """Just take the first values of data along an axis."""
    return np.take(data, 0, axis=axis)


_AGG_FUNCS = {
    "mean": np.mean,
    "median": np.median,
    "min": np.min,
    "max": np.max,
    "sum": np.sum,
    "first": partial(np.take, indices=0),
    "last": partial(np.take, indices=-1),
}


@patch_function()
def aggregate(
    patch: PatchType,
    dim: str,
    method: Literal["mean", "median", "min", "max", "first", "last"] = "mean",
) -> PatchType:
    """
    Aggregate values along a specified dimension.

    Parameters
    ----------
    dim
        The dimension along which aggregations are to be performed.
    method
        The aggregation to apply along dimension. Options are:
            mean, min, max, median, first, last
    """
    axis = patch.dims.index(dim)
    func = _AGG_FUNCS[method]
    new_data = np.expand_dims(func(patch.data, axis=axis), axis)
    # update coords with new dimension (use mean)
    coords = patch.coords
    if dim == "time":  # need to convert time to ints
        ns = patch.coords.get_array(dim).astype(np.int64) / 1_000_000_000
        new_coord_val = to_datetime64(np.mean(ns))
    else:
        new_coord_val = np.mean(patch.coords.get_array(dim))
    new_coords = {
        name: coords.get_array(name) if name != dim else np.array([new_coord_val])
        for name in patch.dims
    }
    return patch.new(data=new_data, coords=new_coords)



"""
Basic support for DAS-HDF5 a subset of CF (climate and forcasting).

This was created mainly for reading PoroTomo data from Brady Hotsprings.

More info on PoroTomo here:
https://github.com/openEDI/documentation/tree/main/PoroTomo
"""
from __future__ import annotations

from .core import DASHDF5



"""Module for waterfall plotting."""
from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np

from dascore.constants import PatchType
from dascore.units import get_quantity_str
from dascore.utils.patch import patch_function
from dascore.utils.plotting import (
    _format_time_axis,
    _get_ax,
    _get_cmap,
    _get_dim_label,
    _get_extents,
)


def _set_scale(im, scale, scale_type, patch):
    """Set the scale of the color bar based on scale and scale_type."""
    # check scale paramters
    assert scale_type in {"absolute", "relative"}
    assert isinstance(scale, float | int) or len(scale) == 2
    # make sure we have a len two array
    data = patch.data
    modifier = 1
    if scale_type == "relative":
        modifier = 0.5 * (np.nanmax(data) - np.nanmin(data))
        # only one scale parameter provided, center around mean
    if isinstance(scale, float):
        mean = np.nanmean(patch.data)
        scale = np.array([mean - scale * modifier, mean + scale * modifier])
    im.set_clim(scale)


@patch_function()
def waterfall(
    patch: PatchType,
    ax: plt.Axes | None = None,
    cmap="bwr",
    scale: float | Sequence[float] | None = None,
    scale_type: Literal["relative", "absolute"] = "relative",
    show=False,
) -> plt.Axes:
    """
    Create a waterfall plot of the Patch data.

    Parameters
    ----------
    patch
        The Patch object.
    ax
        A matplotlib object, if None create one.
    cmap
        A matplotlib colormap string or instance. Set to None to not plot the
        colorbar.
    scale
        If not None, controls the saturation level of the colorbar.
        Values can either be a float, to set upper and lower limit to the same
        value centered around the mean of the data, or a length 2 tuple
        specifying upper and lower limits. See `scale_type` for controlling how
        values are scaled.
    scale_type
        Controls the type of scaling specified by `scale` parameter. Options
        are:
            relative - scale based on half the dynamic range in patch
            absolute - scale based on absolute values provided to `scale`
    show
        If True, show the plot, else just return axis.

    Examples
    --------
    >>> # Plot the default patch
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> _ = patch.viz.waterfall(scale=0.1)
    """
    ax = _get_ax(ax)
    cmap = _get_cmap(cmap)
    data = patch.data
    dims = patch.dims
    assert len(dims) == 2, "Can only make waterfall plot of 2D Patch"
    dims_r = tuple(reversed(dims))
    coords = {dim: patch.coords.get_array(dim) for dim in dims}
    extents = _get_extents(dims_r, coords)
    im = ax.imshow(data, extent=extents, aspect="auto", cmap=cmap, origin="lower")
    # scale colorbar
    if scale is not None:
        _set_scale(im, scale, scale_type, patch)
    for dim, x in zip(dims_r, ["x", "y"]):
        getattr(ax, f"set_{x}label")(_get_dim_label(patch, dim))
        # format all dims which have time types.
        coord = patch.get_coord(dim)
        if np.issubdtype(coord.dtype, np.datetime64):
            _format_time_axis(ax, dim, x)
    # add color bar with title
    if cmap is not None:
        cb = ax.get_figure().colorbar(im, ax=ax, fraction=0.05, pad=0.025)
        data_type = str(patch.attrs["data_type"])
        data_units = get_quantity_str(patch.attrs.data_units) or ""
        dunits = f" ({data_units})" if (data_type and data_units) else f"{data_units}"
        label = f"{data_type}{dunits}"
        cb.set_label(label)
    ax.invert_yaxis()  # invert y axis so origin is at top
    if show:
        plt.show()
    return ax



"""Tests for Trace2D object."""
from __future__ import annotations

import operator
import weakref

import numpy as np
import pandas as pd
import pytest
from rich.text import Text

import dascore as dc
from dascore.core import Patch
from dascore.core.coords import BaseCoord, CoordRange
from dascore.exceptions import CoordError
from dascore.proc.basic import apply_operator


def get_simple_patch() -> Patch:
    """
    Return a small simple array for memory testing.

    Note: This cant be a fixture as pytest seems to hold a reference,
    even for function scoped outputs.
    """
    attrs = {"time_step": 1}
    pa = Patch(
        data=np.random.random((100, 100)),
        coords={"time": np.arange(100) * 0.01, "distance": np.arange(100) * 0.2},
        attrs=attrs,
        dims=("time", "distance"),
    )
    return pa


class TestInit:
    """Tests for init'ing Patch."""

    time1 = np.datetime64("2020-01-01")

    @pytest.fixture()
    def random_dt_coord(self):
        """Create a random patch with a datetime coord."""
        rand = np.random.RandomState(13)
        array = rand.random(size=(20, 200))
        attrs = dict(dx=1, time_step=1 / 250.0, category="DAS", id="test_data1")
        time_deltas = dc.to_timedelta64(np.arange(array.shape[1]) * attrs["time_step"])
        coords = dict(
            distance=np.arange(array.shape[0]) * attrs["dx"],
            time=self.time1 + time_deltas,
        )
        dims = tuple(coords)
        out = dict(data=array, coords=coords, attrs=attrs, dims=dims)
        return Patch(**out)

    @pytest.fixture(scope="class")
    def patch_complex_coords(self):
        """Create a patch with 'complex' (non-dimensional) coords."""
        rand = np.random.RandomState(13)
        array = rand.random(size=(20, 100))
        dt = 1 / 250.0
        attrs = dict(distance_step=1, time_step=dt, category="DAS", id="test_data1")
        time_deltas = dc.to_timedelta64(np.arange(array.shape[1]) * attrs["time_step"])
        coords = dict(
            distance=np.arange(array.shape[0]) * attrs["distance_step"],
            time=self.time1 + time_deltas,
            latitude=("distance", array[:, 0]),
            quality=(("distance", "time"), array),
        )
        dims = ("distance", "time")
        out = dict(data=array, coords=coords, attrs=attrs, dims=dims)
        return Patch(**out)

    @pytest.fixture(scope="class")
    def test_conflicting_attrs_coords_raises(self):
        """Patch for testing conflicting coordinates/attributes."""
        array = np.random.random((10, 10))
        # create attrs, these should all get overwritten by coords.
        attrs = dict(
            distance_step=10,
            time_step=dc.to_timedelta64(1),
            distance_min=1000,
            distance_max=2002,
            time_min=dc.to_datetime64("2017-01-01"),
            time_max=dc.to_datetime64("2019-01-01"),
        )
        # create coords
        coords = dict(
            time=dc.to_datetime64(np.cumsum(np.random.random(10))),
            distance=np.random.random(10),
        )
        # assemble and output.
        dims = ("distance", "time")
        out = dict(data=array, coords=coords, attrs=attrs, dims=dims)
        msg = "Coords and attrs are incompatible."
        with pytest.raises(ValueError, match=msg):
            dc.Patch(**out)

    def test_start_time_inferred_from_dt64_coords(self, random_dt_coord):
        """Ensure the time_min and time_max attrs can be inferred from coord time."""
        patch = random_dt_coord
        assert patch.attrs["time_min"] == self.time1

    def test_end_time_inferred_from_dt64_coords(self, random_dt_coord):
        """Ensure the time_min and time_max attrs can be inferred from coord time."""
        patch = random_dt_coord
        time = patch.coords.coord_map["time"].max()
        assert patch.attrs["time_max"] == time

    def test_init_from_array(self, random_patch):
        """Ensure a trace can be created from raw components; array, coords, attrs."""
        assert isinstance(random_patch, Patch)

    def test_max_time_populated(self, random_patch):
        """Ensure the time_max is populated when not explicitly given."""
        end_time = random_patch.attrs["time_max"]
        assert not pd.isnull(end_time)

    def test_min_max_populated(self, random_patch):
        """The min/max values of the distance attrs should have been populated."""
        attrs = random_patch.attrs
        expected_filled_in = [
            x
            for x in list(attrs.model_fields)
            if x.startswith("distance") and "units" not in x
        ]
        for attr in expected_filled_in:
            assert not pd.isnull(attrs[attr])

    def test_had_default_attrs(self, patch):
        """Test that all patches used in the test suite have default attrs."""
        attr_set = set(patch.attrs.model_dump())
        assert attr_set.issubset(set(patch.attrs.model_dump()))

    def test_no_attrs(self):
        """Ensure a patch with no attrs can be created."""
        pa = Patch(
            data=np.random.random((100, 100)),
            coords={"time": np.random.random(100), "distance": np.random.random(100)},
            dims=("time", "distance"),
        )
        assert isinstance(pa, Patch)

    def test_seconds_as_coords_time_no_dt(self):
        """Ensure seconds passed as coordinates with no attrs still works."""
        ar = get_simple_patch()
        assert not np.any(pd.isnull(ar.coords.get_array("time")))

    def test_shape(self, random_patch):
        """Ensure shape returns the shape of the data array."""
        assert random_patch.shape == random_patch.data.shape

    def test_init_with_complex_coordinates(self, patch_complex_coords):
        """Ensure complex coordinates work."""
        patch = patch_complex_coords
        assert isinstance(patch, Patch)
        assert "latitude" in patch.coords
        assert "quality" in patch.coords
        assert np.all(patch.coords.get_array("quality") == patch.data)

    def test_incomplete_raises(self):
        """An incomplete patch should raise an error."""
        data = np.ones((10, 10))
        with pytest.raises(ValueError, match="data, coords, and dims"):
            Patch(data=data)

    def test_coords_from_1_element_array(self):
        """Ensure CoordRange is still returned despite 1D array in time."""
        patch = dc.get_example_patch("random_das", shape=(100, 1))
        time_coord = patch.get_coord("time")
        assert isinstance(time_coord, CoordRange)

    def test_sin_wave_patch(self):
        """Ensure the sin wave patch is consistent with its coord dims."""
        # For some reason this combination can make coords with wrong shape.
        patch = dc.examples._sin_wave_patch(
            sample_rate=1000,
            frequency=[200, 10],
            channel_count=2,
        )
        assert patch.shape == patch.coords.shape == patch.data.shape
        time_shape = patch.shape[patch.dims.index("time")]
        assert time_shape == len(patch.coords.get_array("time"))
        assert time_shape == len(patch.coords.get_array("time"))

    def test_init_no_coords(self, random_patch):
        """Ensure a new patch can be inited from only attrs."""
        attrs = random_patch.attrs
        new = random_patch.__class__(attrs=attrs, data=random_patch.data)
        assert isinstance(new, dc.Patch)

    def test_init_with_patch(self, random_patch):
        """Ensure a patch inited on a patch just returns a patch."""
        new = dc.Patch(random_patch)
        assert new == random_patch

    def test_non_time_distance_dims(self):
        """Ensure dimensions other than time/distance work."""
        x = np.arange(10) * 2
        y = np.arange(10) * 3
        data = np.add.outer(y, x).astype(np.float64)
        patch = dc.Patch(data=data, coords={"x": x, "y": y}, dims=("x", "y"))
        assert isinstance(patch, dc.Patch)

    def test_patch_has_size(self, random_patch):
        """Ensure patches have same size as data."""
        assert random_patch.size == random_patch.data.size


class TestNew:
    """Tests for `Patch.new` method."""

    def test_erase_history(self, random_patch):
        """Ensure new can erase history."""
        # do some processing, so history shows up.
        patch = random_patch.pass_filter(time=(None, 10)).decimate(time=5)
        assert patch.attrs.history
        new_attrs = dict(patch.attrs)
        new_attrs["history"] = []
        new_patch = patch.new(attrs=new_attrs)
        assert not new_patch.attrs.history

    def test_new_coord_dict_order(self, random_patch):
        """Ensure data can be init'ed with a dict of coords in any orientation."""
        patch = random_patch
        axis = patch.dims.index("time")
        data = np.std(patch.data, axis=axis, keepdims=True)
        new_time = patch.coords.get_array("time")[0:1]
        new_dist = patch.coords.get_array("distance")
        coords_1 = {"time": new_time, "distance": new_dist}
        coords_2 = {"distance": new_dist, "time": new_time}
        # the order the coords are defined shouldn't matter
        out_1 = patch.new(data=data, coords=coords_1)
        out_2 = patch.new(data=data, coords=coords_2)
        assert out_1 == out_2

    def test_attrs_preserved_when_not_specified(self, random_patch):
        """If attrs is not passed to new, old attrs should remain."""
        pa = random_patch.update_attrs(network="bob", tag="2", station="10")
        new_1 = pa.new(data=pa.data * 10)
        assert new_1.attrs == pa.attrs

    def test_new_dims_renames_dims(self, random_patch):
        """Ensure new can rename dimensions."""
        dims = ("tom", "jerry")
        out = random_patch.new(dims=dims)
        assert out.dims == dims


class TestDisplay:
    """Tests for displaying patches."""

    def test_str(self, patch):
        """All patches should have str rep."""
        out = str(patch)
        assert isinstance(out, str)
        assert len(out)

    def test_rich(self, patch):
        """Patches should also have rich representation."""
        out = patch.__rich__()
        assert isinstance(out, Text)

    def test_empty_patch_str(self):
        """An empty patch should have both a str and repr of nonzero length."""
        patch = Patch()
        assert len(str(patch))
        assert len(repr(patch))


class TestEmptyPatch:
    """Tests for empty patch objects."""

    def test_no_dims(self):
        """An empty test patch should have no dims or related attrs."""
        patch = Patch()
        attrs = patch.attrs
        assert len(attrs)
        assert not len(patch.dims)


class TestEquals:
    """Tests for checking equality."""

    def test_equal_self(self, random_patch):
        """Ensure a trace equals itself."""
        assert random_patch.equals(random_patch)

    def test_non_equal_array(self, random_patch):
        """Ensure the traces are not equal if the data are not equal."""
        new_data = random_patch.data + 1
        new = random_patch.new(data=new_data)
        assert not new.equals(random_patch)

    def test_coords_named_differently(self, random_patch):
        """Ensure if the coords are named differently patches are not equal."""
        dims = random_patch.dims
        new_coords = dict(random_patch.coords)
        new_coords["bob"] = new_coords.pop(dims[-1])
        new_dims = tuple(list(dims)[:-1] + ["bob"])
        patch_2 = random_patch.new(coords=new_coords, dims=new_dims)
        assert not patch_2.equals(random_patch)

    def test_coords_not_equal(self, random_patch):
        """Ensure if the coords are not equal neither are the arrays."""
        new_coords = dict(random_patch.coords)
        new_coords["distance"] = new_coords["distance"] + 10
        patch_2 = random_patch.new(coords=new_coords)
        assert not patch_2.equals(random_patch)

    def test_attrs_not_equal(self, random_patch):
        """Ensure if the attributes are not equal the arrays are not equal."""
        attrs = random_patch.attrs
        new_attr_dict = {}
        new_attr_dict["time_step"] = attrs["time_step"] * 0.80
        patch2 = random_patch.new(attrs=new_attr_dict)
        assert patch2.attrs.time_step == new_attr_dict["time_step"]
        assert not patch2.equals(random_patch)

    def test_one_null_value_in_attrs(self, random_patch):
        """Ensure setting a value to null in attrs doesn't eval to equal."""
        attrs = dict(random_patch.attrs)
        attrs["tag"] = ""
        patch2 = random_patch.new(attrs=attrs)
        patch2.equals(random_patch)
        assert not patch2.equals(random_patch)

    def test_extra_attrs(self, random_patch):
        """
        Ensure extra attrs in one patch still eval equal unless only_required
        is used.
        """
        attrs = dict(random_patch.attrs)
        attrs["new_label"] = "fun4eva"
        patch2 = random_patch.new(attrs=attrs)
        # they should be equal
        assert patch2.equals(random_patch)
        # until extra labels are allowed
        assert not patch2.equals(random_patch, only_required_attrs=False)

    def test_both_attrs_nan(self, random_patch):
        """If Both Attrs are some type of NaN the patches should be equal."""
        attrs1 = dict(random_patch.attrs)
        attrs1["label"] = np.NaN
        patch1 = random_patch.new(attrs=attrs1)
        attrs2 = dict(attrs1)
        attrs2["label"] = None
        patch2 = random_patch.new(attrs=attrs2)
        assert patch1.equals(patch2)

    def test_transposed_patches_not_equal(self, random_patch):
        """Transposed patches are not considered equal."""
        transposed = random_patch.transpose()
        # make sure dims are NE
        assert transposed.attrs.dims != random_patch.attrs.dims
        assert transposed.dims != random_patch.dims
        # and test equality
        assert not random_patch.equals(transposed)

    def test_one_coord_not_equal(self, wacky_dim_patch):
        """Ensure coords being close but not equals fails equality."""
        patch = wacky_dim_patch
        coords = patch.coords
        coord_array = np.array(coords.coord_map["distance"].values)
        coord_array[20:30] *= 0.9
        assert not np.allclose(coord_array, coords.get_array("distance"))
        new_patch = patch.update_coords(distance=coord_array)
        new = patch.new(coords=new_patch.coords)
        assert new != patch

    def test_other_types(self, random_patch):
        """Ensure a non-patch is not equal."""
        assert not random_patch.equals(None)
        assert not random_patch.equals(1)
        assert not random_patch.equals(random_patch.data)


class TestTranspose:
    """Tests for switching dimensions."""

    def test_simple_transpose(self, random_patch):
        """Just check dims and data shape are reversed."""
        dims = random_patch.dims
        dims_r = tuple(reversed(dims))
        pa1 = random_patch
        pa2 = pa1.transpose(*dims_r)
        assert pa2.dims == dims_r
        assert list(pa2.data.shape) == list(reversed(pa1.data.shape))


class TestUpdateAttrs:
    """
    Tests for updating the attrs.

    Note: Most of the testing for this functionality is actually done on
    dascore.util.patch.AttrsCoordsMixer.
    """

    def test_add_attr(self, random_patch):
        """Tests for adding an attribute."""
        new = random_patch.update_attrs(bob=1)
        assert "bob" in new.attrs.model_dump() and new.attrs["bob"] == 1

    def test_original_unchanged(self, random_patch):
        """Updating attributes shouldn't change original patch in any way."""
        old_attrs = random_patch.attrs.model_dump()
        _ = random_patch.update_attrs(bob=2)
        assert not hasattr(random_patch.attrs, "bob")
        assert random_patch.attrs.model_dump() == old_attrs

    def test_update_starttime1(self, random_patch):
        """Ensure coords are updated with attrs."""
        t1 = np.datetime64("2000-01-01")
        pa = random_patch.update_attrs(time_min=t1)
        assert pa.attrs["time_min"] == t1
        assert pa.coords.min("time") == t1

    def test_update_startttime2(self, random_patch):
        """Updating start time should update end time as well."""
        duration = random_patch.attrs["time_max"] - random_patch.attrs["time_min"]
        new_start = np.datetime64("2000-01-01")
        pa1 = random_patch.update_attrs(time_min=str(new_start))
        assert pa1.attrs["time_min"] == new_start
        assert pa1.attrs["time_max"] == new_start + duration

    def test_dt_is_datetime64(self, random_patch):
        """Ensure dt gets changed into timedelta64."""
        d_time = random_patch.attrs["time_step"]
        assert isinstance(d_time, np.timedelta64)
        new1 = random_patch.update_attrs(time_step=10)
        assert new1.attrs["time_step"] == dc.to_timedelta64(10)

    def test_update_non_sorted_coord(self, wacky_dim_patch):
        """Ensure update attrs updates non-sorted coordinates."""
        # test updating dist max
        pa = wacky_dim_patch.update_attrs(distance_max=10)
        assert pa.attrs.distance_max == 10
        assert not np.any(pd.isnull(pa.coords.get_array("distance")))
        # test update dist min
        pa = wacky_dim_patch.update_attrs(distance_min=10)
        assert pa.attrs.distance_min == 10
        assert not np.any(pd.isnull(pa.coords.get_array("distance")))

    def test_update_units(self, random_patch):
        """Ensure units can be updated in attrs."""
        new_dist = "ft"
        patch1 = random_patch.update_attrs(distance_units=new_dist)
        patch2 = random_patch.convert_units(distance=new_dist)
        coord1 = patch1.get_coord("distance")
        coord2 = patch2.get_coord("distance")
        coord3 = random_patch.get_coord("distance").convert_units(new_dist)
        assert coord1 == coord2 == coord3
        assert patch1 == patch2


class TestReleaseMemory:
    """Ensure memory is released when the patch is deleted."""

    def test_single_patch(self):
        """Ensure a single patch is gc'ed when it leaves scope."""
        simple_patch = get_simple_patch()
        wr = weakref.ref(simple_patch.data)
        # delete pa and ensure the array was collected.
        del simple_patch
        assert wr() is None

    def test_decimated_patch(self):
        """
        Ensure a process which changes the array lets the first array get gc'ed.

        This is very important since the main reason to decimate is to preserve
        memory.
        """
        patch = get_simple_patch()
        new = patch.decimate(time=10)
        wr = weakref.ref(patch.data)
        del patch
        assert isinstance(new, Patch)
        assert wr() is None

    def test_select(self):
        """A similar test to ensure select releases old memory."""
        patch = get_simple_patch()
        new = patch.select(time=[0.1, 10], copy=True)
        wr = weakref.ref(patch.data)
        del patch
        assert isinstance(new, Patch)
        assert wr() is None


class TestPipe:
    """Tests for piping Patch to other functions."""

    @staticmethod
    def pipe_func(patch, positional_arg, keyword_arg=None):
        """Just add passed arguments to stats, return new patch."""
        out = patch.update_attrs(positional_arg=positional_arg, keyword_arg=keyword_arg)
        return out

    def test_pipe_basic(self, random_patch):
        """Test simple application of pipe."""
        out = random_patch.pipe(self.pipe_func, 2)
        assert out.attrs["positional_arg"] == 2
        assert out.attrs["keyword_arg"] is None

    def test_keyword_arg(self, random_patch):
        """Ensure passing a keyword arg works."""
        out = random_patch.pipe(self.pipe_func, 2, keyword_arg="bob")
        assert out.attrs["positional_arg"] == 2
        assert out.attrs["keyword_arg"] == "bob"


class TestCoords:
    """Test various things about patch coordinates."""

    @pytest.fixture(scope="class")
    def random_patch_with_lat(self, random_patch):
        """Create a random patch with added lat/lon coordinates."""
        dist = random_patch.coords.get_array("distance")
        lat = np.arange(0, len(dist)) * 0.001 - 109.857952
        # add a single coord
        out = random_patch.update_coords(latitude=("distance", lat))
        return out

    def test_add_single_dim_one_coord(self, random_patch_with_lat):
        """Tests that one coordinate can be added to a patch."""
        assert "latitude" in random_patch_with_lat.coords

    def test_add_single_dim_two_coord2(self, random_patch_with_lat_lon):
        """Ensure multiple coords can be added to patch."""
        out2 = random_patch_with_lat_lon
        assert {"latitude", "longitude"}.issubset(set(out2.coords.coord_map))
        assert out2.coords.get_array("longitude").shape
        assert out2.coords.get_array("latitude").shape

    def test_add_multi_dim_coords(self, multi_dim_coords_patch):
        """Ensure coords with multiple dimensions works."""
        out1 = multi_dim_coords_patch
        assert "quality" in out1.coords
        assert out1.coords.get_array("quality").shape
        assert np.all(out1.coords.get_array("quality") == 1)

    def test_coord_time_narrow_select(self, multi_dim_coords_patch):
        """Ensure the coord type doesn't change in narrow slice."""
        patch = multi_dim_coords_patch
        time = patch.coords.coord_map["time"]
        new = patch.select(time=(time.min(), time.min()))
        assert 1 in new.shape
        new_coords = new.coords.coord_map
        assert isinstance(new_coords["time"], CoordRange)


class TestApplyOperator:
    """Tests for applying various ufunc-type operators."""

    ops = (
        operator.add,
        operator.sub,
        operator.mul,
        operator.truediv,
        operator.floordiv,
        operator.pow,
    )
    bool_ops = (
        operator.and_,
        operator.or_,
    )

    @pytest.fixture(params=ops)
    def operation(self, request):
        """Parameterization to get operator."""
        return request.param

    def test_scalar1(self, random_patch, operation):
        """Tests for scalar operators."""
        new = operation(random_patch, 2)
        expected = operation(random_patch.data, 2)
        assert np.allclose(new.data, expected)

    def test_array_like(self, random_patch):
        """Ensure array-like operations work."""
        ones = np.ones(random_patch.shape)
        new = apply_operator(random_patch, ones, np.add)
        assert np.allclose(new.data, ones + random_patch.data)


class TestGetCoord:
    """Tests for retrieving coords and imposing requirements."""

    def test_simple(self, random_patch):
        """Ensure a simple coordinate is returned."""
        coord = random_patch.get_coord("time")
        assert coord == random_patch.coords.coord_map["time"]

    def test_raises_not_evenly_sampled(self, wacky_dim_patch):
        """Ensure an error is raised when coord is not evenly sampled, if specified."""
        patch = wacky_dim_patch
        assert isinstance(patch.get_coord("time"), BaseCoord)
        match = "Coordinate time is not evenly sampled"
        with pytest.raises(CoordError, match=match):
            patch.get_coord("time", require_evenly_sampled=True)

    def test_raises_not_sorted(self, wacky_dim_patch):
        """Ensure an error is raised when coord is not sorted, if specified."""
        patch = wacky_dim_patch
        assert isinstance(patch.get_coord("distance"), BaseCoord)
        match = "Coordinate distance is not sorted"
        with pytest.raises(CoordError, match=match):
            patch.get_coord("distance", require_sorted=True)


class TestAssertHasCoords:
    """Test suite for has coords."""

    def test_raises_single(self, random_patch):
        """Ensure an error is raised if as single required coord isnt found."""
        match = "does not have required coordinate"
        with pytest.raises(CoordError, match=match):
            random_patch.assert_has_coords("foo")

    def test_raises_multiple(self, random_patch):
        """Ensure an error is raised if required coords aren't found."""
        match = "does not have required coordinate"
        with pytest.raises(CoordError, match=match):
            random_patch.assert_has_coords(("bar", "depth"))

    def test_has_coords_ok(self, random_patch):
        """No Error should be raised if the patch has the coordinates."""
        dims = random_patch.dims
        random_patch.assert_has_coords(dims[0])
        random_patch.assert_has_coords(dims)


class TestDeprecations:
    """Ensure deprecations are issued."""

    def test_assign_coords_deprecated(self, random_patch):
        """assign_coords should issue dep. warning."""
        new_coord = random_patch.coords.get_array("time")

        with pytest.warns(DeprecationWarning):
            random_patch.assign_coords(new_time=("time", new_coord))


class TestSetDims:
    """Tests for setting dimensions."""

    # note: patch.set_dims just passes work down too CoordManager.set_dims
    # which is tested more thoroughly
    def test_set_dim(self, random_patch_with_lat_lon):
        """Simple test for set_dims."""
        patch = random_patch_with_lat_lon
        out = patch.set_dims(distance="longitude")
        assert "longitude" in out.dims
        assert "longitude" in out.coords.dim_map["distance"]

    def test_doctest_example(self, random_patch):
        """Ensure the doctest example works."""
        patch = random_patch
        my_coord = np.random.random(patch.coord_shapes["time"])
        out = patch.update_coords(my_coord=("time", my_coord)).set_dims(  # add my_coord
            time="my_coord"
        )  # set mycoord as dim (rather than time)
        assert "my_coord" in out.dims



"""Utility for working with time."""
from __future__ import annotations

from datetime import datetime, timedelta
from functools import singledispatch

import numpy as np
import pandas as pd

from dascore.constants import (
    LARGEDT64,
    NUMPY_TIME_UNIT_MAPPING,
    ONE_SECOND,
    SMALLDT64,
    timeable_types,
)


@singledispatch
def to_datetime64(obj: timeable_types | np.ndarray):
    """
    Convert an object to a datetime64.

    This function accepts a wide range of inputs and returns something
    of the same shape, but converted to numpy's datetime64 representation.

    Examples
    --------
    >>> # Convert an [iso 8601](https://en.wikipedia.org/wiki/ISO_8601) string
    >>> import dascore as dc
    >>> time = dc.to_datetime64('2017-09-17T12:11:01.23212')
    >>> # Convert a timestamp (float)
    >>> dt = dc.to_datetime64(631152000.0)
    """
    if pd.isnull(obj):
        return np.datetime64("NaT")
    msg = f"type {type(obj)} is not supported"
    raise NotImplementedError(msg)


@to_datetime64.register(str)
def _str_to_datetime64(obj: str) -> np.datetime64:
    """Convert a string to a datetime64 object."""
    return np.datetime64(obj, "ns")


@to_datetime64.register(float)
@to_datetime64.register(int)
@to_datetime64.register(np.number)
def _float_to_datetime(num: float | int) -> np.datetime64:
    """Convert a float to a single datetime."""
    ar = np.array([num])
    return _array_to_datetime64(ar)[0]


@to_datetime64.register(np.ndarray)
@to_datetime64.register(list)
@to_datetime64.register(tuple)
def _array_to_datetime64(array: np.ndarray) -> np.datetime64 | np.ndarray:
    """Convert an array of floating point timestamps to an array of np.datatime64."""
    array = np.array(array)
    nans = pd.isnull(array)
    # dealing with objects
    if np.issubdtype(array.dtype, np.dtype(object)):
        array = np.array([to_datetime64(x) for x in array]).astype("datetime64[ns]")
    # dealing with a string
    if np.issubdtype(array.dtype, np.dtype(str)):
        array = array.astype("datetime64[ns]")
    # dealing with an array of datetime64 or empty array
    if np.issubdtype(array.dtype, np.datetime64) or len(array) == 0:
        if not array.shape:  # dealing with degenerate (0-D( array
            out = np.datetime64(array)
        else:
            out = array.astype("datetime64[ns]")
    # dealing with numerical data
    elif not np.issubdtype(array.dtype, np.datetime64) and np.isreal(array[0]):
        with np.errstate(divide="ignore", invalid="ignore"):
            array[nans] = 0  # temporary replace NaNs
            abs_array = np.abs(array)
            sign = np.sign(array)
            # separate seconds and factions, assume ns precision
            int_sec = abs_array.astype(np.int64)
            frac_sec = abs_array % 1.0
            ns = (frac_sec * 1_000_000_000).astype(np.int64)
            out = (sign * (int_sec * 1_000_000_000 + ns)).astype("datetime64[ns]")
        # fill NaN Back in
        out[nans] = np.datetime64("NaT")
    return out


@to_datetime64.register(pd.Series)
def _float_to_datetime(ser: pd.Series) -> pd.Series:
    """Convert a float to a single datetime."""
    ar = to_datetime64(ser.values)
    return pd.Series(ar, index=ser.index)


@to_datetime64.register(np.datetime64)
def _pass_datetime(datetime):
    """Simply return the datetime."""
    return np.datetime64(datetime, "ns")


@to_datetime64.register(datetime)
def _datetime_to_datetime64(dt: datetime):
    """Convert python datetime to datetime64."""
    # because pandas NaT has datettime in its MRO we need to check
    # if this is nullish and return NaT if so.
    if pd.isnull(dt):
        return np.datetime64("NaT")
    return to_datetime64(np.datetime64(dt))


@to_datetime64.register(pd.Timestamp)
def _pandas_timestamp(datetime: pd.Timestamp):
    return datetime.to_datetime64()


@singledispatch
def to_timedelta64(obj: float | np.ndarray | str | timedelta):
    """
    Convert an object to timedelta64.

    This function accepts a wide range of inputs and returns something
    of the same shape, but converted to numpy's timedelta64 representation.

    Examples
    --------
    >>> # Convert a float to seconds
    >>> import dascore as dc
    >>> d_time_1 = dc.to_timedelta64(10.1232)
    >>> # also works on negative numbers
    >>> d_time_2 = dc.to_datetime64(-10.5)
    """
    if pd.isnull(obj):
        return np.timedelta64("NaT")
    msg = f"type {type(obj)} is not supported"
    raise NotImplementedError(msg)


@to_timedelta64.register(float)
@to_timedelta64.register(int)
@to_timedelta64.register(np.number)
def _float_to_timedelta64(num: float | int) -> np.datetime64:
    """Convert a float to a single datetime."""
    ar = np.array([num])
    return _array_to_timedelta64(ar)[0]


@to_timedelta64.register(np.ndarray)
@to_timedelta64.register(list)
@to_timedelta64.register(tuple)
def _array_to_timedelta64(array: np.array) -> np.datetime64:
    """Convert an array of floating point timestamps to an array of np.datatime64."""
    array = np.array(array)
    nans = pd.isnull(array)
    array[nans] = 0
    # convert pure object arrays into float so sign casting works.
    if np.issubdtype(array.dtype, np.dtype(object)):
        array = array.astype(np.float64)
    if np.issubdtype(array.dtype, np.timedelta64) or len(array) == 0:
        if not array.shape:  # unpack degenerate array
            return np.timedelta64(array)
        else:
            return array.astype("timedelta64[ns]")
    assert np.isreal(array[0])
    # inf/NaN complain, salience these types of warnings for this block.
    with np.errstate(divide="ignore", invalid="ignore"):
        # separate seconds and factions, convert fractions to ns precision,
        # sub in array. Track sign and use abs to ensure sign comes out.
        sign = np.sign(array)
        abs_array = np.abs(array)
        seconds = abs_array.astype(np.int64).astype("timedelta64[s]")
        frac_sec = abs_array % 1.0
        ns = (frac_sec * 1_000_000_000).astype(np.int64).astype("timedelta64[ns]")
        out = sign * (seconds + ns)
        out[nans] = np.timedelta64("NaT")
    return out


@to_timedelta64.register(pd.Series)
def _series_to_timedelta64_series(ser: pd.Series) -> pd.Series:
    """Convert a series to a series of timedelta64."""
    out = to_timedelta64(ser.values)
    return pd.Series(out, index=ser.index)


@to_timedelta64.register(np.timedelta64)
def _pass_time_delta(time_delta):
    """Simply return the time delta as ns precision."""
    return time_delta.astype("<m8[ns]")


@to_timedelta64.register(pd.Timedelta)
def _unpack_pandas_time_delta(time_delta: pd.Timedelta):
    """Simply return the time delta."""
    return time_delta.to_numpy()


@to_timedelta64.register(timedelta)
def _timedelta_to_timedelta64(td):
    """Return timedelta64."""
    return to_timedelta64(np.timedelta64(td))


@to_timedelta64.register(str)
def _time_delta_from_str(time_delta_str: str):
    """Simply return the time delta."""
    split = time_delta_str.split(" ")
    assert len(split) == 2
    val, units = split
    if units[-1] == "s":
        units = units[:-1]
    new_unit = NUMPY_TIME_UNIT_MAPPING[units]
    return np.timedelta64(int(val), new_unit)


@singledispatch
def to_int(obj: timeable_types | np.ndarray) -> np.ndarray:
    """
    Ensure a scalar or array is a number.

    If the input values represents a time or a time-delta, convert it to a
    an int representing ns.
    """
    msg = f"type {type(obj)} is not supported"
    raise NotImplementedError(msg)


@to_int.register(float)
@to_int.register(int)
@to_int.register(np.number)
def _float_to_num(num: float | int) -> float | int:
    """Convert number to int."""
    return int(num)


@to_int.register(np.ndarray)
@to_int.register(list)
@to_int.register(tuple)
def _array_to_int(array: np.ndarray) -> np.ndarray:
    """Convert an array of floating point timestamps to an array of np.datatime64."""
    array = np.array(array)
    if not len(array):
        return array.astype(np.int64)
    # dealing with an array of datetime64 or empty array
    is_dt = np.issubdtype(array.dtype, np.datetime64)
    is_td = np.issubdtype(array.dtype, np.timedelta64)
    if is_td or is_dt:
        new = to_datetime64(array) if is_dt else to_timedelta64(array)
        array = new.astype(np.int64)
    return array


@to_int.register(np.datetime64)
@to_int.register(datetime)
@to_int.register(pd.Timestamp)
def _time_to_int(datetime):
    """Simply return the datetime converted to ns."""
    return to_int([to_datetime64(datetime)])[0]


@to_int.register(type(None))
@to_int.register(type(pd.NaT))
@to_int.register(type(pd.NA))
def _return_number_null(null):
    """Convert non to NaT."""
    return np.NaN


@to_int.register(np.timedelta64)
def _time_delta_to_number(time_delta: np.timedelta64):
    return to_int([to_timedelta64(time_delta)])[0]


@to_int.register(pd.Series)
def _pandas_timestamp_to_num(ser: pd.Series):
    return ser.view(np.int64)


@singledispatch
def to_float(obj: timeable_types | np.ndarray) -> np.ndarray:
    """
    Convert various datetime/timedelta things to a float.

    Time offsets represent seconds, and datetimes are seconds from 1970.
    """
    return float(obj)


@to_float.register(np.ndarray)
@to_float.register(list)
@to_float.register(tuple)
def _array_to_float(array: np.ndarray) -> np.ndarray:
    """Convert an array of floating point timestamps to an array of np.datatime64."""
    array = np.array(array)
    if not len(array):
        return array.astype(np.float64)
    if np.issubdtype(array.dtype, np.datetime64):
        # convert to offset from 1970
        array = array - to_datetime64(0)
    if np.issubdtype(array.dtype, np.timedelta64):
        array = array / ONE_SECOND
    return array.astype(np.float64)


@to_float.register(np.datetime64)
@to_float.register(datetime)
@to_float.register(pd.Timestamp)
def _time_to_float(datetime):
    """Simply return the datetime."""
    td = to_datetime64(datetime) - to_datetime64(0)
    return to_float(td)


@to_float.register(type(None))
@to_float.register(type(pd.NaT))
@to_float.register(type(pd.NA))
def _return_null(null):
    """Convert non to NaT."""
    return np.NaN


@to_float.register(np.timedelta64)
@to_float.register(timedelta)
@to_float.register(pd.Timedelta)
def _time_delta_to_float(time_delta: np.timedelta64):
    return to_timedelta64(time_delta) / ONE_SECOND


def is_datetime64(obj) -> bool:
    """Return True if object is a datetime64 (or equivalent) or an array of such."""
    if isinstance(obj, np.datetime64):
        return True
    if isinstance(obj, np.ndarray | list | tuple | pd.Series):
        if np.issubdtype(np.array(obj).dtype, np.datetime64):
            return True
    if isinstance(obj, pd.Timestamp):
        return True
    return False


def is_timedelta64(obj) -> bool:
    """Return True if object is a timedelta64 (or equivalent) or an array of such."""
    if isinstance(obj, np.timedelta64):
        return True
    if isinstance(obj, np.ndarray | list | tuple | pd.Series):
        if np.issubdtype(np.array(obj).dtype, np.timedelta64):
            return True
    if isinstance(obj, pd.Timedelta):
        return True
    return False


def dtype_time_like(dtype_or_array) -> bool:
    """Return True if dtype is time related (datetime64, timedelta64)."""
    try:
        dtype_or_array = np.dtype(dtype_or_array)
    except TypeError:
        dtype_or_array = getattr(dtype_or_array, "dtype", dtype_or_array)
    is_datetime = np.issubdtype(dtype_or_array, np.datetime64)
    is_timedelta = np.issubdtype(dtype_or_array, np.timedelta64)
    if is_timedelta or is_datetime:
        return True
    return False


def get_max_min_times(kwarg_time=None):
    """
    Function to get min/max times from a tuple of possible time values.

    If None, return max/min times possible.
    """
    # first unpack time from tuples
    assert kwarg_time is None or len(kwarg_time) == 2
    time_min, time_max = (None, None) if kwarg_time is None else kwarg_time
    # get defaults if starttime or endtime is none
    time_min = None if pd.isnull(time_min) else time_min
    time_max = None if pd.isnull(time_max) else time_max
    time_min = to_datetime64(time_min or SMALLDT64)
    time_max = to_datetime64(time_max or LARGEDT64)
    if time_min is not None and time_max is not None:
        if time_min > time_max:
            msg = "time_min cannot be greater than time_max."
            raise ValueError(msg)
    return time_min, time_max



"""
Module for managing coordinates.

The coordmanager is a simple class for tracking multidimensional coordinates
and labels of ndarrays. Various methods exist for trimming, sorting, and
filtering the managed arrays based on coordinates.

# Initializing CoordinateManagers

Coordinate managers are initialized using the
[get_coords](`dascore.core.coordmanager.get_coord_manager`) function. They take
a combination of coordinate dictionaries, dimensions, and attributes. They can
also be extracted from example patches.

```{python}
import dascore as dc
from rich import print

patch = dc.get_example_patch()
data = patch.data
cm = patch.coords
print(cm)
```

```{python}
import numpy as np
# Get array of coordinate values
time = cm['time']

# Filter data array
# this gets a new coord manager and a view of the trimmed data.
_sorted_time = np.sort(time)
t1 = _sorted_time[5]
t2 = _sorted_time[-5]
new_cm, new_data = cm.select(time=(t1, t2), array=data)
print(f"Old data shape: {data.shape}")
print(f"New data shape: {new_data.shape}")
print(new_cm)
```
"""
from __future__ import annotations

import warnings
from collections import defaultdict
from collections.abc import Mapping, Sequence, Sized
from functools import reduce
from operator import and_, or_
from typing import Annotated, Any, TypeVar

import numpy as np
from pydantic import field_validator, model_validator
from rich.text import Text
from typing_extensions import Self

import dascore as dc
from dascore.constants import dascore_styles
from dascore.core.coords import BaseCoord, CoordSummary, get_coord
from dascore.exceptions import (
    CoordDataError,
    CoordError,
    CoordMergeError,
    CoordSortError,
    ParameterError,
)
from dascore.utils.display import get_nice_text
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import (
    _matches_prefix_suffix,
    all_close,
    cached_method,
    iterate,
    separate_coord_info,
)
from dascore.utils.models import (
    ArrayLike,
    DascoreBaseModel,
    frozen_dict_serializer,
    frozen_dict_validator,
)

MaybeArray = TypeVar("MaybeArray", ArrayLike, np.ndarray, None)


def _validate_select_coords(coord, coord_name):
    """Ensure multi-dims are not used."""
    if not len(coord.shape) == 1:
        msg = (
            "Only 1 dimensional coordinates can be used for selection "
            f"{coord_name} has {len(coord.shape)} dimensions."
        )
        raise CoordError(msg)


def _indirect_coord_updates(cm, dim_name, coord_name, reduction, new_coords):
    """
    Applies trim to coordinates.

    Assumes other associated coordinates are trimmed.
    """
    other_coords = set(cm.dim_to_coord_map[dim_name]) - {coord_name}
    # perform indirect updates.
    for icoord in other_coords:
        dims, coord = new_coords[icoord]
        axis = cm.dim_map[icoord].index(dim_name)
        new = coord.index(reduction, axis=axis)
        new_coords[icoord] = (dims, new)


def _to_slice(limits):
    """Convert slice or two len tuple to slice."""
    if isinstance(limits, slice):
        return limits
    # ints should be interpreted as Slice(int, int+1) to not collapse dim.
    if isinstance(limits, int):
        if limits == -1:  # -1 case needs open interval to work
            return slice(-1, None)
        return slice(limits, limits + 1)
    if limits is ... or limits is None:
        return slice(None, None)
    assert isinstance(limits, Sized) and len(limits) == 2
    val1, val2 = limits
    start = None if val1 is ... or val1 is None else val1
    stop = None if val2 is ... or val2 is None else val2
    return slice(start, stop)


def _get_indexers_and_new_coords_dict(cm, kwargs, samples=False, relative=False):
    """Function to get reductions for each dimension."""
    dim_reductions = {x: slice(None, None) for x in cm.dims}
    dimap = cm.dim_map
    new_coords = dict(cm._get_dim_array_dict(keep_coord=True))
    for coord_name, limits in kwargs.items():
        # this is not a selectable coord, just skip.
        if coord_name not in cm.coord_map or not len(cm.dim_map[coord_name]):
            continue
        coord = cm.coord_map[coord_name]
        _validate_select_coords(coord, coord_name)
        dim_name = dimap[coord_name][0]
        # different logic if we are using indices or values
        if not samples:
            new_coord, reductions = coord.select(limits, relative=relative)
        else:
            reductions = _to_slice(limits)
            new_coord = coord[reductions]
        # this handles the case of out-of-bound selections.
        # These should be converted to degenerate coords.
        dim_reductions[dim_name] = reductions
        new_coords[coord_name] = (dimap[coord_name], new_coord)
        # update other coords affected by change.
        _indirect_coord_updates(cm, dim_name, coord_name, reductions, new_coords)
    indexers = tuple(dim_reductions[x] for x in cm.dims)
    return new_coords, indexers


class CoordManager(DascoreBaseModel):
    """
    Class for managing coordinates.

    Attributes
    ----------
    dims
        A tuple of dimension names.
    coord_map
        A mapping of {coord_name: Coord}
    dim_map
        A mapping of {coord_name: (dimensions, ...)}
    """

    dims: tuple[str, ...]
    coord_map: Annotated[
        FrozenDict[str, BaseCoord],
        frozen_dict_validator,
        frozen_dict_serializer,
    ]
    dim_map: Annotated[
        FrozenDict[str, tuple[str, ...]],
        frozen_dict_validator,
        frozen_dict_serializer,
    ]

    @model_validator(mode="before")
    @classmethod
    def _validate_coords(cls, values):
        """Validate the coordinates and dimensions."""
        coord_map, dim_map = values["coord_map"], values["dim_map"]
        dims = values["dims"]
        try:
            dim_shapes = {dim: coord_map[dim].shape for dim in dims}
        except KeyError:
            missing = set(dims) - set(coord_map)
            msg = f"All dimensions must have coordinates, {missing} are missing."
            raise CoordError(msg)
        # ensure non-dimensional coordinates have the same length as
        # corresponding coordinates.
        for name, coord_dims in dim_map.items():
            expected_shape = tuple(dim_shapes[x][0] for x in coord_dims)
            shape = coord_map[name].shape
            if tuple(expected_shape) == shape or not expected_shape:
                continue
            msg = (
                f"coordinate: {name} has a shape of {shape} which does not "
                f"match the dimension(s) of {coord_dims} which have a shape "
                f"of {expected_shape}"
            )
            raise CoordError(msg)
        # Ensure dimensional coordinates are associated as such
        for dim in dims:
            assert dim_map[dim] == (dim,), f"Incorrect dim association on {dim}"

        return values

    @field_validator("coord_map", "dim_map")
    @classmethod
    def _convert_to_frozen_dicts(cls, v):
        """Ensure mapping fields are immutable."""
        return FrozenDict(v)

    def __getitem__(self, item) -> np.ndarray:
        # in order to not break backward compatibility, we need to return
        # the array. Start the deprecation cycle to be consistent.
        msg = (
            "Currently coords[coord_name] returns a numpy array, but in a "
            "future dascore version it will return a BaseCoord instance. "
            "Use patch.coords.get_array(coord_name) instead."
        )
        warnings.warn(msg, UserWarning, stacklevel=2)
        return self._get_coord(item).values

    def __getattr__(self, item) -> BaseCoord:
        try:
            return super().__getattr__(item)
        except AttributeError:
            # unlike get item, get attr returns the base coordinate.
            return self._get_coord(item)

    def __iter__(self):
        for key in self.coord_map:
            yield (key, self.get_array(key))

    def __contains__(self, key):
        return key in self.coord_map

    def update_coords(self, **kwargs) -> Self:
        """
        Update the coordinates, return a new Coordinate Manager.

        Input values can be of the same form as initialization.
        To drop coordinates, simply pass {coord_name: None}
        """

        def _get_dim_change_drop(coord_map, dim_map):
            """
            Determine which coords must be dropped because their corresponding
            dimension size changed.
            """
            out = []
            for i, _v in coord_map.items():
                if i not in self.dims:
                    continue
                assert len(dim_map[i]) == 1, "only dealing with 1D dimensions"
                if len(coord_map[i]) == len(self.coord_map[i]):
                    continue
                for coord_name, dims in self.dim_map.items():
                    if i in dims:
                        out.append(coord_name)
            return out

        def _divide_kwargs(kwargs):
            """Divide kwargs into coords to update, drop, add, etc."""
            prefix = self.dims
            suffix = ("step", "min", "max")
            coord_updates = {
                x: kwargs.pop(x)
                for x in set(kwargs)
                if _matches_prefix_suffix(x, prefixes=prefix, suffixes=suffix)
            }
            coords_to_drop = [i for i, v in kwargs.items() if v is None]
            # convert input to coord_map/dim_map
            coords_to_add = {i: v for i, v in kwargs.items() if v is not None}
            return coord_updates, coords_to_drop, coords_to_add

        coord_updates, coord_to_drop, coord_to_add = _divide_kwargs(kwargs)
        # get coords to drop from selecting None
        coord_map, dim_map = _get_coord_dim_map(coord_to_add, self.dims)
        # find coords to drop because their dimension changed.
        indirect_coord_drops = _get_dim_change_drop(coord_map, dim_map)
        # drop coords then call get_coords to handle adding new ones.
        coords, _ = self.drop_coords(*(coord_to_drop + indirect_coord_drops))
        out = coords._get_dim_array_dict()
        out.update({i: v for i, v in kwargs.items() if i not in coord_to_drop})
        # update based on keywords
        for item, value in coord_updates.items():
            coord_name, attr = item.split("_")
            new = list(out[coord_name])
            coord = get_coord(values=new[1])
            new[1] = coord.update(**{attr: value})
            out[coord_name] = tuple(new)
        dims = tuple(x for x in self.dims if x not in coord_to_drop)
        return get_coord_manager(out, dims=dims)

    def update_from_attrs(
        self, attrs: Mapping | dc.PatchAttrs
    ) -> tuple[Self, dc.PatchAttrs]:
        """
        Update coordinates from attrs.

        This will also return a PatchAttrs which conforms to coords.

        Parameters
        ----------
        attrs
            The attribute source, either PatchAttrs instance or mapping.
        """
        coord_info, attr_info = separate_coord_info(attrs, dims=self.dims)
        out = dict(self.coord_map)
        for name in set(coord_info) & set(out):
            maybe_updates = coord_info[name]
            coord = self.coord_map[name]
            # convert values to dict to determine which should be updated.
            model_contents = coord.to_summary().model_dump(exclude_defaults=True)
            # see what has changed
            diff = {
                i: v for i, v in maybe_updates.items() if v != model_contents.get(i)
            }
            out[name] = coord.update(**diff)
        coords = self.new(coord_map=out)
        # anything not used in coord_info should be put back.
        # for example, data_units might get put in its own coord called data.
        for unused_dim in set(coord_info) - set(coords.dims):
            for key, val in coord_info[unused_dim].items():
                attr_info[f"{unused_dim}_{key}"] = val
        attr_info["coords"] = coords.to_summary_dict()
        attr_info["dims"] = coords.dims
        attrs = dc.PatchAttrs.from_dict(attr_info)
        return coords, attrs

    def sort(
        self, *coords, array: MaybeArray = None, reverse: bool = False
    ) -> tuple[Self, MaybeArray]:
        """
        Sort coordinates.

        Parameters
        ----------
        *coords
            Specify which dimensions to sort. If None, sort along dimensional
            coordinates.
        array
            The array to sort.
        reverse
            If true, sort in descending order, else ascending.

        Raises
        ------
        CoordSortError if sorting is not possible.
        """

        def _validate_coords(coords):
            """Coordinates should be one dimensional."""
            dims = [self.dim_map[x] for x in coords]
            if any(len(x) > 1 for x in dims):
                bad_coord = [y for x, y in zip(dims, coords, strict=True) if len(y) > 1]
                msg = (
                    f"cannot simultaneously sort coordinate(s) {bad_coord}"
                    f" because they are associated with more than one dimension."
                )
                raise CoordSortError(msg)
            if len(coords) == 1:  # no need to check conflict if one dim
                return
            flat_dims = [x for y in dims for x in y]
            if len(flat_dims) != len(set(flat_dims)):
                msg = f"cannot merge {coords} because they share a dimension"
                raise CoordSortError(msg)

        def _get_dimensional_sorts(coords2sort):
            """Create the indexer for each coord."""
            new_coords = dict(self.coord_map)
            indexes = []
            for name, coord in coords2sort.items():
                dim = self.dim_map[name][0]
                new_coord, indexer = coord.sort(reverse=reverse)
                new_coords[name] = new_coord
                indexes.append(self._get_indexer(self.dims.index(dim), indexer))
                # also sort related coords.
                _sort_related(name, dim, indexer, new_coords)
            return new_coords, tuple(indexes)

        def _sort_related(name, dim, indexer, new_coords):
            """Find and apply sorting to related coords."""
            related_coords = set(self.dim_to_coord_map[dim]) - {name}
            for r_coord_name in related_coords:
                coord = new_coords[r_coord_name]
                ind = self.dim_map[r_coord_name].index(dim)
                slicer = [slice(None, None) for _ in coord.shape]
                slicer[ind] = indexer
                new_coords[r_coord_name] = coord[tuple(slicer)]

        # get coords that need sorting
        attr = "reverse_sorted" if reverse else "sorted"
        coords = coords if coords else self.dims
        _validate_coords(coords)
        cmap = self.coord_map
        assert set(coords).issubset(set(cmap))
        coords2sort = {x: cmap[x] for x in coords if not getattr(cmap[x], attr)}
        new_coords, indexers = _get_dimensional_sorts(coords2sort)
        if array is not None:
            for index in indexers:
                array = array[index]
        out = self.new(coord_map=new_coords)
        assert out.shape == self.shape
        return out, array

    def snap(
        self, *coords, array: MaybeArray = None, reverse: bool = False
    ) -> tuple[Self, MaybeArray]:
        """
        Force the specified coordinates to be monotonic and evenly sampled.

        Coordinates are first sorted, then snapped.

        Parameters
        ----------
        *coords
            Specify which dimensions to sort. If None, sort along dimensional
            coordinates.
        array
            The array to sort/snap.
        reverse
            If true, sort in descending order, else ascending.
        """
        coords = self.dims if len(coords) == 0 else coords
        cm, array = self.sort(*coords, array=array, reverse=reverse)
        # now the arrays are sorted it should be correct to snap dimensions.
        cmap = dict(cm.coord_map)
        for coord_name in coords:
            cmap[coord_name] = cmap[coord_name].snap()
        out = cm.new(coord_map=cmap)
        assert out.shape == self.shape
        return out, array

    def new(self, dims=None, coord_map=None, dim_map=None) -> Self:
        """
        Return a new coordmanager with specified attributes replaced.

        Parameters
        ----------
        dims
            A tuple of dimension strings.
        coord_map
            A mapping of {name: coord}
        dim_map
            A mapping of {coord_name:
        """
        out = self.__class__(
            dims=dims if dims is not None else self.dims,
            coord_map=coord_map if coord_map is not None else self.coord_map,
            dim_map=dim_map if dim_map is not None else self.dim_map,
        )
        return out

    def drop_coords(
        self,
        *coords: str | Sequence[str],
        array: MaybeArray = None,
    ) -> tuple[Self, MaybeArray]:
        """
        Drop one or more coordinates.

        If the coordinate is a dimension, also drop other coords that depend
        on that dimension.

        Parameters
        ----------
        *coords
            The name of the coordinate or dimension.
        """
        dim_drop_list = []
        coords_to_drop = {x for x in iterate(coords)}
        # If there are either no coords to drop or this cm doesn't have them.
        if not coords_to_drop or not (set(self.coord_map) & coords_to_drop):
            return self, array
        # figure out which coords to drop are dimensional coordinates.
        for name, dims in self.dim_map.items():
            if coords_to_drop & set(dims):
                dim_drop_list.append(name)
        dims_to_drop = set(dim_drop_list)
        # combine dims and plain coordinates that should be dropped.
        total_drops = dims_to_drop | coords_to_drop
        coord_map = {i: v for i, v in self.coord_map.items() if i not in total_drops}
        dim_map = {i: v for i, v in self.dim_map.items() if i not in total_drops}
        dims = tuple(x for x in self.dims if x not in dims_to_drop)
        index = tuple(
            slice(None, None) if x not in coords_to_drop else slice(0, 0)
            for x in self.dims
        )
        new = self.__class__(coord_map=coord_map, dim_map=dim_map, dims=dims)
        return new, self._get_new_data(index, array)

    def disassociate_coord(self, *coord: str) -> Self:
        """
        Disassociate some coordinates from dimensions.

        These coordinates will no longer be associated with a dimension in
        the coord manager but can still be retrieved.

        Parameters
        ----------
        coord
            The coordinate name(s) to disassociated from their dimensions.
        """
        new = {x: (None, self.coord_map[x]) for x in coord}
        return self.drop_coords(*coord)[0].update_coords(**new)

    def drop_disassociated_coords(self) -> Self:
        """Drop all coordinates not associated with a dimension."""
        cmap = self.coord_map
        dim_map = self.dim_map
        no_dim_coords = [x for x in cmap if dim_map[x] == ()]
        return self.drop_coords(*no_dim_coords)

    def set_dims(self, **kwargs: str) -> Self:
        """
        Set dimension to non-dimensional coordinate.

        Parameters
        ----------
        kwargs
            A mapping indicating old_dim: new_dim where new_dim refers to
            the name of a non-dimensional coordinate.
        """
        dims = list(self.dims)
        coord_map = dict(self.coord_map)
        dim_map = dict(self.dim_map)
        for old_dim, new_dim in kwargs.items():
            if new_dim not in coord_map or old_dim not in dims:
                msg = (
                    f"{old_dim} is not a dimension or {new_dim} is not a "
                    f"coordinate."
                )
                raise CoordError(msg)
            # ensure coords have the same shape
            old_coord, new_coord = coord_map[old_dim], coord_map[new_dim]
            if not old_coord.shape == new_coord.shape or len(new_coord.shape) != 1:
                msg = (
                    f"New coordinate {new_coord} with a shape of {new_coord.shape} "
                    f"does not match the shape of {old_coord} ({old_coord.shape})"
                )
                raise CoordError(msg)

            # swap out old dims
            dims[dims.index(old_dim)] = new_dim
        # now loop over dim_maps and swap out old dims with new
        old_to_new = {i: v for i, v in zip(self.dims, dims, strict=True)}
        for coord_name, coord_dims in dim_map.items():
            dim_map[coord_name] = tuple(old_to_new[x] for x in coord_dims)
        return self.__class__(dims=dims, coord_map=coord_map, dim_map=dim_map)

    def select(
        self, array: MaybeArray = None, relative=False, samples=False, **kwargs
    ) -> tuple[Self, MaybeArray]:
        """
        Perform value-based selection on coordinates.

        Parameters
        ----------
        array
            An array to which the selection will be applied.
        relative
            If True, coordinate updates are relative.
        samples
            If True, the query meaning is in samples.
        **kwargs
            Used to specify select arguments. Can be of the form
            {coord_name: (lower_limit, upper_limit)}.
        """
        new_coords, indexers = _get_indexers_and_new_coords_dict(
            self, kwargs, samples=samples, relative=relative
        )
        new_cm = self.update_coords(**new_coords)
        return new_cm, self._get_new_data(indexers, array)

    def _get_new_data(self, indexer, array: MaybeArray) -> MaybeArray:
        """Get new data array after applying some trimming."""
        if array is None:  # no array passed, just return.
            return array
        return array[indexer]

    def __rich__(self) -> str:
        """Rich formatting for the coordinate manager."""
        dc_blue = dascore_styles["dc_blue"]
        header_text = Text("➤ ") + Text("Coordinates", style=dc_blue) + Text(" (")
        lens = {x: self.coord_map[x].shape[0] for x in self.dims}
        dim_texts = Text(", ").join(
            [
                Text(x, style="bold")
                + Text(": ")
                + Text(f"{lens[x]}", style="underline")
                for x in self.dims
            ]
        )
        out = [header_text, dim_texts, ")"]
        # sort coords by dims, coords
        non_dim_coords = sorted(set(self.coord_map) - set(self.dims))
        names = list(self.dims) + non_dim_coords
        for name in names:
            coord = self.coord_map[name]
            coord_dims = self.dim_map[name]
            if name in self.dims:
                base = Text.assemble("\n    * ", Text(name, style="bold"), ": ")
            else:
                base = Text(f"\n    {name} {coord_dims}: ")
            text = Text.assemble(base, coord.__rich__())
            out.append(text)
        return Text.assemble(*out)

    def __str__(self):
        return str(self.__rich__())

    def equals(self, other) -> bool:
        """Return True if other coordinates are approx equal."""
        if not isinstance(other, self.__class__):
            return False
        if not set(self.coord_map) == set(other.coord_map):
            return False
        coord_1, coord_2 = self.coord_map, other.coord_map
        for name, _coord in self.coord_map.items():
            if not all_close(coord_1[name].values, coord_2[name].values):
                return False
        return True

    __eq__ = equals

    @property
    def shape(self):
        """Return the shape of the dimensions."""
        out = tuple(len(self.coord_map[x]) for x in self.dims)
        # empty arrays return (0,) as their shape, so we must do the same.
        if not out:
            return (0,)
        return out

    @property
    def size(self):
        """Return the shape of the dimensions."""
        return np.prod(self.shape)

    def validate_data(self, data):
        """Ensure data conforms to coordinates."""
        data = np.array([]) if data is None else data
        if self.shape != data.shape:
            msg = (
                f"Data array has a shape of {data.shape} which doesnt match "
                f"the coordinate manager shape of {self.shape}."
            )
            raise CoordDataError(msg)
        return data

    def _get_dim_array_dict(self, keep_coord=False):
        """
        Get the coord map in the form:
        {coord_name = ((dims,), array)}.

        if keep_coord, just keep the coordinate as second arg.
        """
        out = {}
        for name, coord in self.coord_map.items():
            dims = self.dim_map[name]
            out[name] = (dims, coord if keep_coord else coord.data)
        return out

    def set_units(self, **kwargs):
        """Set the units of the coordinate manager."""
        new_coords = dict(self.coord_map)
        for name, units in kwargs.items():
            new_coords[name] = new_coords[name].set_units(units)
        return self.new(coord_map=new_coords)

    def convert_units(self, **kwargs):
        """
        Convert units in coords according to kwargs. Will raise if incompatible
        Coordinates are specified.
        """
        new_coords = dict(self.coord_map)
        for name, units in kwargs.items():
            new_coords[name] = new_coords[name].convert_units(units)
        return self.new(coord_map=new_coords)

    def simplify_units(self):
        """Simplify all units in the coordinates."""
        new_coords = dict(self.coord_map)
        for name, coord in new_coords.items():
            new_coords[name] = coord.simplify_units()
        return self.new(coord_map=new_coords)

    def transpose(self, *dims: str | type(Ellipsis)) -> Self:
        """Transpose the coordinates."""

        def _get_transpose_dims(new, old):
            """Get a full tuple of dimensions for transpose, allowing for ..."""
            new_set = set(new)
            new_list, old_list = list(new), list(old)
            if len(new_set) != len(new):
                msg = "Transpose cannot use duplicate dimensions or ... more than once."
                raise ParameterError(msg)
            # iterate new list, pop out values from old list.
            for value in new_set - {...}:
                old_list.remove(value)
            # sub in remaining values for ...
            if ... in new_set:
                ind = new_list.index(...)
                new_list = new_list[:ind] + old_list + new_list[ind + 1 :]
            # ensure all values are accounted for
            if set(new_list) != set(old):
                msg = (
                    "You must specify all dimensions in a transpose operation "
                    f"or use ... You passed {new} but dimensions are {old}"
                )
                raise ParameterError(msg)
            return tuple(new_list)

        dims = _get_transpose_dims(new=dims or self.dims[::-1], old=self.dims)
        return self.new(dims=dims)

    def rename_coord(self, **kwargs) -> Self:
        """
        Rename the coordinates or dimensions.

        Parameters
        ----------
        **kwargs
            Used to rename coord {old_coord_name: new_coord_name}
        """

        def _adjust_dim_map_to_new_dim(dim_map, old_name, new_name):
            """Get the updated dim map."""
            out = dim_map
            # replace all mentions of
            for key, value in out.items():
                if old_name not in value:
                    continue
                new_value = tuple(x if x != old_name else new_name for x in value)
                out[key] = new_value

            return out

        dims, coord_map = list(self.dims), dict(self.coord_map)
        dim_map = dict(self.dim_map)
        for old_name, new_name in kwargs.items():
            if old_name == new_name:
                continue
            if old_name in dims:  # only
                dims[dims.index(old_name)] = new_name
                dim_map = _adjust_dim_map_to_new_dim(dim_map, old_name, new_name)
            coord_map[new_name] = coord_map.pop(old_name)
            dim_map[new_name] = dim_map.pop(old_name)

        out = dict(dims=dims, coord_map=coord_map, dim_map=dim_map)
        return self.__class__(**out)

    def squeeze(self, dim: Sequence[str] | None = None) -> Self:
        """
        Squeeze length one dimensions.

        Parameters
        ----------
        dim
            The dimension name, or sequence of dimension names, to drop.
            If None, drop all length 1 dims

        Raises
        ------
        CoordError if the selected dimension has a length gt 1.
        """
        to_drop = []
        if dim is None:
            dim = [x for x in self.dims if len(self.coord_map[x]) <= 1]
        bad_dim = set(iterate(dim)) - set(self.dims)
        if bad_dim:
            msg = (
                f"The following dims cannot be dropped because they don't "
                f"exist in the coordinate manager {bad_dim}"
            )
            raise CoordError(msg)
        for name in iterate(dim):
            coord = self.coord_map[name]
            if len(coord) > 1:
                msg = f"cant squeeze dim {name} because it has non-zero length"
                raise CoordError(msg)
            to_drop.append(name)
        return self.drop_coords(*to_drop)[0]

    def decimate(self, **kwargs) -> tuple[Self, tuple[slice, ...]]:
        """
        Evenly subsample along some dimension.

        Parameters
        ----------
        **kwargs
            Used to specify the dimension to decimate.

        Notes
        -----
        Removes any coordinate which depended on the decimated dimension.
        """
        assert len(kwargs) == 1
        (dim, value) = next(iter(kwargs.items()))
        assert dim in self.dims
        dim_slice = slice(None, None, int(value))
        new_array = self.coord_map[dim][dim_slice]
        new = self.update_coords(**{dim: new_array})
        slices = tuple(
            slice(None, None) if d != dim else slice(None, None, value)
            for d in new.dims
        )
        return new, slices

    @property
    @cached_method
    def coord_shapes(self) -> dict[str, tuple[int, ...]]:
        """Get a dict of {coord_name: shape}."""
        return {i: v.shape for i, v in self.coord_map.items()}

    @property
    @cached_method
    def dim_to_coord_map(self) -> FrozenDict[str, tuple[str, ...]]:
        """Get a dimension to coordinate map."""
        out = defaultdict(list)
        for coord, dims in self.dim_map.items():
            for dim in dims:
                out[dim].append(coord)
        return FrozenDict({i: tuple(v) for i, v in out.items()})

    def get_coord_tuple_map(self) -> dict[str, tuple[tuple[str, ...], BaseCoord]]:
        """
        Return a mapping of {coord_name: (dims, coord)}.

        This is helpful because it can be passed directly to
        get_coordinate_manager.
        """
        dims = self.dim_map
        coords = self.coord_map
        return {x: (dims[x], c) for x, c in coords.items()}

    def _get_coord_dims_tuple(self):
        """Return a tuple of ((coord, dims...,), ...)."""
        dim_map = self.dim_map
        return tuple((name, *dim_map[name]) for name in self.coord_map)

    def _get_indexer(self, ind: int | None = None, value=None):
        """
        Get an indexer for the appropriate data shape.

        This is useful for generating a tuple that can be used

        ind is a list of indices to substitute in values.
        """
        out = [slice(None, None) for _ in self.shape]
        out[ind] = value
        return tuple(out)

    def to_summary_dict(self) -> dict[str, CoordSummary]:
        """Convert the contents of the coordinate manager to a summary dict."""
        dim_map = self.dim_map
        out = {}
        for name, coord in self.coord_map.items():
            out[name] = coord.to_summary(dims=dim_map[name])
        return out

    def _get_coord(self, coord_name):
        """Get a coord. Raise an error if the coordinate is not in the coordmanager."""
        if coord_name not in self.coord_map:
            msg = (
                f"No coordinate named {coord_name} in coord manager. "
                f"Valid coordinates are {list(self.coord_map)}."
            )
            raise AttributeError(msg)
        return self.coord_map[coord_name]

    def min(self, coord_name):
        """Return the minimum value of a coordinate."""
        return self._get_coord(coord_name).min()

    def max(self, coord_name):
        """Return the maximum value of a coordinate."""
        return self._get_coord(coord_name).max()

    def step(self, coord_name):
        """Return the coordinate step."""
        return self._get_coord(coord_name).step

    def get_array(self, coord_name) -> np.ndarray:
        """Return the coordinate values as a numpy array."""
        return np.array(self._get_coord(coord_name))


def get_coord_manager(
    coords: Mapping[str, BaseCoord | np.ndarray] | CoordManager | None = None,
    dims: tuple[str, ...] | None = None,
    attrs: dc.PatchAttrs | dict[str, Any] | None = None,
) -> CoordManager:
    """
    Create a coordinate manager.

    Parameters
    ----------
    coords
        Information about coordinates. These can be a mapping of the
        form: {name, array}, {name: (dim_name, array)}, or
        {name: ((dim_names,) array). Can also be a
        [`CoordManager`](`dascore.core.CoordManager`).
    dims
        Tuple specify dimension names
    attrs
        Attributes which can be used to create coordinates.
        Cannot be used with coords argument.
        If you want to update [`CoordManager`](`dascore.core.CoordManager`)
        use [`update_from_attrs`](`dascore.core.CoordManager.update_from_attrs`).

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> from dascore.core.coordmanager import get_coord_manager
    >>> # initialize coordinates from dict of arrays
    >>> distance = np.arange(0, 100)
    >>> time = dc.to_datetime64(np.arange(0, 100_000, 1_000))
    >>> coords = {"distance": distance, "time": time}
    >>> dims = ("time", "distance")
    >>> cm = get_coord_manager(coords=coords, dims=dims)
    >>> # add non-dimension 1D coordinate. Note the coord dict must be a tuple
    >>> # with the first element specifying the dimensional coord the
    >>> # non-dimensional coord is attached to.
    >>> latitude = np.random.random(distance.shape)
    >>> coords['latitude'] = ('distance', latitude)
    >>> cm = get_coord_manager(coords=coords, dims=dims)
    >>> # Add two-D non-dimensional coord.
    >>> quality = np.random.random((len(distance), len(time)))
    >>> coords['quality'] = (("distance", "time"), quality)
    >>> cm = get_coord_manager(coords=coords, dims=dims)
    >>> # Get coordinate manager from typical patch attribute dict
    >>> attrs = dc.get_example_patch().attrs
    >>> cm = get_coord_manager(attrs=attrs)
    """
    if coords is not None and attrs is not None:
        msg = (
            "Cannot use both attrs and coords in get_coord_manager. "
            "Perhaps you want CoordManager.update_from_attrs?"
        )
        raise ParameterError(msg)
    # return coords if we already have a coord manager.
    if isinstance(coords, CoordManager):
        # maybe try to rename dims.
        if dims is not None and dims != coords.dims:
            kwargs = {i: v for i, v in zip(coords.dims, dims, strict=True)}
            coords = coords.rename_coord(**kwargs)
        return coords
    # this allows a simple dict without dims to be passed and dims pulled
    # from dict keys.
    if dims is None:
        if isinstance(coords, Mapping) and all(isinstance(x, str) for x in coords):
            dims = tuple(coords.keys())
        elif attrs is not None and "dims" in attrs or hasattr(attrs, "dims"):
            dims = tuple(attrs["dims"].split(","))
        else:
            dims = ()
    coords = {} if coords is None else coords
    coord_map, dim_map = _get_coord_dim_map(coords, dims)
    if attrs:
        coord_updates, _ = separate_coord_info(attrs, dims)
        updateable_coords = set(coord_updates) - set(coord_map)
        for name in updateable_coords:
            coord_map[name] = get_coord(**coord_updates[name])
            dim_map[name] = (name,)
    out = CoordManager(coord_map=coord_map, dim_map=dim_map, dims=dims)
    return out


def _get_coord_dim_map(coords, dims):
    """Get coord_map and dim_map from coord input."""

    def _get_coord(coord):
        """Get a coordinate from various inputs."""
        if hasattr(coord, "model_dump"):
            coord = coord.model_dump()
        if isinstance(coord, Mapping):  # input is a dict
            out = get_coord(**coord)
        else:
            out = get_coord(values=coord)
        return out

    def _coord_from_simple(name, coord):
        """Get coordinates from {coord_name: coord} where coord_name is dim name."""
        if name not in dims:
            msg = (
                "Coordinates that are not named the same as dimensions "
                "must be passed as a tuple of the form: "
                "(dimension, coord) "
            )
            raise CoordError(msg)
        out = _get_coord(coord)
        return out, (name,)

    def _maybe_coord_from_nested(coord):
        """
        Get coordinates from {coord_name: (dim_name, coord)} or
        {coord_name: ((dim_names...,), coord)}.
        """
        if not len(coord) == 2:
            msg = (
                "Second input for coords must be length two of the form:"
                " (dimension, coord) or ((dimensions,...), coord)"
            )
            raise CoordError(msg)
        dim_names = iterate(coord[0])
        # all dims must be in the input dims.
        if not (d1 := set(dim_names)).issubset(d2 := set(dims)):
            bad_dims = d2 - d1
            msg = (
                f"Coordinate specified invalid dimension(s) {bad_dims}."
                f" Valid dimensions are {dims}"
            )
            raise CoordError(msg)
        # pull out any relevant info from attrs.
        coord_out = _get_coord(coord[1])
        assert coord_out.shape == np.shape(coord[1])
        return coord_out, dim_names

    assert not isinstance(coords, CoordManager)
    # this is a dead code path, but may be needed later, leaving for now.
    # if isinstance(coords, CoordManager):
    #     coords_dump = coords.model_dump()
    #     return dict(coords_dump["coord_map"]), dict(coords_dump["dim_map"])

    c_map, d_map = {}, {}
    # iterate coords, get coordinate output.
    for name, coord in coords.items():
        if not isinstance(coord, tuple):
            c_map[name], d_map[name] = _coord_from_simple(name, coord)
        else:
            c_map[name], d_map[name] = _maybe_coord_from_nested(coord)
    return c_map, d_map


def merge_coord_managers(
    coord_managers: Sequence[CoordManager],
    dim: str,
    snap_tolerance: float | None = None,
) -> CoordManager:
    """
    Merger coordinate managers along a specified dimension.

    Parameters
    ----------
    coord_managers
        A sequence of coord_managers to merge.
    dim
        The dimension along which to merge.
    snap_tolerance
        The tolerance for snapping CoordRanges together. E.G, allows
        coord ranges that have snap_tolerances differences from their
        start/end to be joined together. If they don't meet this requirement
        an [CoordMergeError](`dascore.exceptions.CoordMergeError`) is raised.
        If None, no checks are performed.
    """

    def _get_dims(managers):
        """Ensure all managers have same dimensions."""
        dims = {x.dims for x in managers}
        if len(dims) != 1:
            msg = (
                "Can't merge coord managers, they don't all have the "
                "same dimensions!"
            )
            raise CoordMergeError(msg)
        return managers[0].dims

    def _drop_unshared_coordinates(managers):
        """Any coordinates not shared between managers should be dropped."""
        # gets [{(coord, dims, ...), (coord, dims, ...)}, ...] to ensure
        # both the coords name and their dimensions are common between managers
        coord_sets = [set(x._get_coord_dims_tuple()) for x in managers]
        common_coords = reduce(and_, coord_sets)
        all_coords = reduce(or_, coord_sets)
        if not (drop_coords := all_coords - common_coords):
            return managers
        coords_to_drop = [x[0] for x in drop_coords]
        return [x.drop_coords(*coords_to_drop)[0] for x in managers]

    def _get_non_merge_coords(managers, non_merger_names):
        """Ensure all non-merge coords are equal."""
        out = {}
        for coord_name in non_merger_names:
            first = managers[0].coord_map[coord_name]
            if all([first == x.coord_map[coord_name] for x in managers]):
                dims = managers[0].dim_map[coord_name]
                out[coord_name] = (dims, first)
                continue
            msg = (
                f"Non merging coordinates {coord_name} are not equal. "
                "Coordinate managers cannot be merged."
            )
            raise CoordMergeError(msg)
        return out

    def _snap_coords(coord_list):
        """Snap coordinates together."""
        if snap_tolerance is None:
            return coord_list  # skip snapping if no snap tolerance.
        for ind in range(1, len(coord_list)):
            c_coord = coord_list[ind - 1]
            n_coord = coord_list[ind]
            tolerance = snap_tolerance * c_coord.step
            assumed_start = c_coord.max() + c_coord.step
            diff = np.abs(assumed_start - n_coord.min())
            # snap is close enough, update coord.
            if diff > 0 and diff <= tolerance:
                coord_list[ind] = n_coord.update_limits(min=assumed_start)
            # snap is too far off, bail out.
            elif diff > tolerance:
                msg = (
                    f"Cannot merge. Snap tolerance: {get_nice_text(tolerance)}"
                    f" not met"
                )
                raise CoordMergeError(msg)
        return coord_list

    def _get_merged_coords(managers, coords_to_merge):
        """Get the merged coordinates."""
        out = {}
        for coord_name in coords_to_merge:
            merge_coords = [x.coord_map[dim] for x in managers]
            axis = managers[0].dim_map[coord_name].index(dim)
            if len(units := {x.units for x in merge_coords}) != 1:
                # TODO: we might try to convert all the units to a common
                # unit in the future.
                msg = (
                    f"Cannot merge coordinates {coord_name}, they dont all "
                    f"share the same units. Units found are: {set(units)}"
                )
                raise CoordMergeError(msg)
            snap_coords = _snap_coords(merge_coords)
            datas = [x.data for x in snap_coords]
            dims = managers[0].dim_map[dim]
            new_data = np.concatenate(datas, axis=axis)
            out[coord_name] = (dims, new_data)
        return out

    def _get_new_coords(managers) -> dict[str, tuple[tuple[str, ...], ArrayLike]]:
        """Merge relevant coordinates together."""
        # build up merged coords.
        coords_to_merge = managers[0].dim_to_coord_map[dim]
        coords_not_to_merge = set(managers[0].coord_map) - set(coords_to_merge)
        # non-merging coordinates should be identical.
        coords_dict = _get_non_merge_coords(managers, coords_not_to_merge)
        # merge coordinates
        coords_dict.update(_get_merged_coords(managers, coords_to_merge))
        return coords_dict

    dims = _get_dims(coord_managers)
    coord_managers = _drop_unshared_coordinates(coord_managers)
    sort_managers = sorted(coord_managers, key=lambda x: x.coord_map[dim].min())
    merged_coords = _get_new_coords(sort_managers)
    return get_coord_manager(merged_coords, dims=dims)



"""A 2D trace object."""
from __future__ import annotations

import warnings
from collections.abc import Mapping, Sequence

import numpy as np
from rich.text import Text
from typing_extensions import Self

import dascore as dc
import dascore.proc.coords
import dascore.utils.io
from dascore import transform
from dascore.compat import DataArray, array
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import CoordManager, get_coord_manager
from dascore.core.coords import BaseCoord
from dascore.utils.display import array_to_text, attrs_to_text, get_dascore_text
from dascore.utils.models import ArrayLike
from dascore.viz import VizPatchNameSpace


class Patch:
    """
    A Class for managing data and metadata.

    Parameters
    ----------
    data
        The array data representing fiber optic measurements.
    coords
        The coordinates, or dimensional labels for the data.
        A few types of input are permitted. If a mapping (eg dict) the value
        should conform to one of the following three forms:
        {coord_name: coord}
        {coord_name: ((dimensions,), coord)}
        {coord_name: (dimensions, coord)}
        Where coord can be a numpy array or a
        [`BaseCoord`](`dascore.core.coords.BaseCoord`) object.
        A [`CoordManager`](`dascore.core.coordmanager.CoordManager`) is also acceptable.
    dims
        A sequence of dimension strings. The first entry corresponds to the
        first axis of data, the second to the second dimension, and so on.
    attrs
        Optional attributes (non-coordinate metadata) passed as a dict or
        [PatchAttrs](`dascore.core.attrs.PatchAttrs')

    Notes
    -----
    - If coordinates and dims are not provided, they will be extracted from
    attrs, if possible.

    - If coords and attrs are provided, attrs will have priority. This means
    if there is a conflict between information contained in both, the coords
    will be recalculated.
    """

    data: ArrayLike
    coords: CoordManager
    dims: tuple[str, ...]
    attrs: PatchAttrs | Mapping

    def __init__(
        self,
        data: ArrayLike | DataArray | None = None,
        coords: Mapping[str, ArrayLike | BaseCoord] | CoordManager | None = None,
        dims: Sequence[str] | None = None,
        attrs: Mapping | PatchAttrs | None = None,
    ):
        if isinstance(data, DataArray | self.__class__):
            data, attrs, coords = data.data, data.attrs, data.coords
        if dims is None and isinstance(coords, CoordManager):
            dims = coords.dims
        # Try to generate coords from ranges in attrs
        if coords is None and attrs is not None:
            attrs = dc.PatchAttrs.from_dict(attrs)
            coords = attrs.coords_from_dims()
            dims = dims if dims is not None else attrs.dim_tuple
        # Ensure required info is here
        non_attrs = [x is None for x in [data, coords, dims]]
        if any(non_attrs) and not all(non_attrs):
            msg = "data, coords, and dims must be defined to init Patch."
            raise ValueError(msg)
        coords = get_coord_manager(coords, dims=dims)
        # the only case we allow attrs to include coords is if they are both
        # dicts, in which case attrs might have unit info for coords.
        if isinstance(attrs, Mapping) and attrs:
            coords, attrs = coords.update_from_attrs(attrs)
        else:
            # ensure attrs conforms to coords
            attrs = dc.PatchAttrs.from_dict(attrs).update(coords=coords)
        self._coords = coords
        self._attrs = attrs
        self._data = array(self.coords.validate_data(data))

    def __eq__(self, other):
        """Compare one Patch."""
        return dascore.proc.equals(self, other)

    def __add__(self, other):
        return dascore.proc.apply_operator(self, other, np.add)

    def __sub__(self, other):
        return dascore.proc.apply_operator(self, other, np.subtract)

    def __floordiv__(self, other):
        return dascore.proc.apply_operator(self, other, np.floor_divide)

    def __truediv__(self, other):
        return dascore.proc.apply_operator(self, other, np.divide)

    def __mul__(self, other):
        return dascore.proc.apply_operator(self, other, np.multiply)

    def __pow__(self, other):
        return dascore.proc.apply_operator(self, other, np.power)

    def __rich__(self):
        dascore_text = get_dascore_text()
        patch_text = Text("Patch ⚡", style="bold")
        header = Text.assemble(dascore_text, " ", patch_text)
        line = Text("-" * len(header))
        coords = self.coords.__rich__()
        attrs = self.attrs
        data = array_to_text(self.data, units=attrs.get("data_units"))
        attrs = attrs_to_text(self.attrs)
        out = Text("\n").join([header, line, coords, data, attrs])
        return out

        pass

    def __str__(self):
        out = self.__rich__()
        return str(out)

    __repr__ = __str__

    @property
    def dims(self) -> tuple[str, ...]:
        """Return the dimensions contained in patch."""
        return self.coords.dims

    @property
    def coord_shapes(self) -> dict[str, tuple[int, ...]]:
        """Return a dict of coordinate: (shape, ...)."""
        return self.coords.coord_shapes

    @property
    def attrs(self) -> PatchAttrs:
        """Return the dimensions contained in patch."""
        return self._attrs

    @property
    def coords(self) -> CoordManager:
        """Return the dimensions contained in patch."""
        return self._coords

    @property
    def data(self) -> ArrayLike:
        """Return the dimensions contained in patch."""
        return self._data

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the data array."""
        return self.coords.shape

    @property
    def size(self) -> tuple[int, ...]:
        """Return the shape of the data array."""
        return self.coords.size

    # --- basic patch functionality.

    new = dascore.proc.new
    equals = dascore.proc.equals
    update_attrs = dascore.proc.update_attrs
    assert_has_coords = dascore.proc.assert_has_coords
    get_coord = dascore.proc.get_coord
    pipe = dascore.proc.pipe
    set_dims = dascore.proc.set_dims
    squeeze = dascore.proc.squeeze
    transpose = dascore.proc.transpose
    snap_coords = dascore.proc.snap_coords
    sort_coords = dascore.proc.sort_coords
    rename_coords = dascore.proc.rename_coords
    update_coords = dascore.proc.update_coords
    drop_coords = dascore.proc.drop_coords
    coords_from_df = dascore.proc.coords_from_df

    def assign_coords(self, *args, **kwargs):
        """Deprecated method for update_coords."""
        msg = "assign_coords is deprecated, use update_coords instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.update_coords(*args, **kwargs)

    set_units = dascore.proc.set_units
    convert_units = dascore.proc.convert_units
    simplify_units = dascore.proc.simplify_units

    # --- processing funcs

    select = dascore.proc.select

    def iselect(self, *args, **kwargs):
        """Deprecated  form of select."""
        msg = "patch.iselect is deprecated. Use patch.select with samples=True"
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.select(*args, samples=True, **kwargs)

    correlate = dascore.proc.correlate
    decimate = dascore.proc.decimate
    detrend = dascore.proc.detrend
    dropna = dascore.proc.dropna
    pass_filter = dascore.proc.pass_filter
    sobel_filter = dascore.proc.sobel_filter
    median_filter = dascore.proc.median_filter
    aggregate = dascore.proc.aggregate
    abs = dascore.proc.abs
    real = dascore.proc.real
    imag = dascore.proc.imag
    angle = dascore.proc.angle
    resample = dascore.proc.resample

    def iresample(self, *args, **kwargs):
        """Deprecated method."""
        msg = (
            "patch.iresample is deprecated. Please use patch.resample "
            "with samples=True"
        )
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.resample(*args, samples=True, **kwargs)

    interpolate = dascore.proc.interpolate
    normalize = dascore.proc.normalize
    standardize = dascore.proc.standardize
    taper = dascore.proc.taper
    rolling = dascore.proc.rolling
    whiten = dascore.proc.whiten

    # --- transformation functions
    differentiate = transform.differentiate
    rfft = transform.rfft
    dft = transform.dft
    idft = transform.idft
    integrate = transform.integrate
    spectrogram = transform.spectrogram
    velocity_to_strain_rate = transform.velocity_to_strain_rate
    dispersion_phase_shift = transform.dispersion_phase_shift

    # --- Method Namespaces
    # Note: these can't be cached_property (from functools) or references
    # to self stick around and keep large arrays in memory.

    @property
    def viz(self) -> VizPatchNameSpace:
        """The visualization namespace."""
        return VizPatchNameSpace(self)

    @property
    def tran(self) -> Self:
        """The transformation namespace."""
        msg = (
            "The tran namespace is deprecated. Its methods can now be "
            "accessed as normal path methods (eg patch.dft)"
        )
        warnings.warn(msg, DeprecationWarning)
        return self

    @property
    def io(self) -> dc.io.PatchIO:
        """Return a patch IO object for saving patches to various formats."""
        return dc.io.PatchIO(self)



"""Tests module for wave format."""
from __future__ import annotations

from pathlib import Path

import pytest
from scipy.io.wavfile import read as read_wav

import dascore as dc


class TestWriteWav:
    """Tests for writing wav format to disk."""

    @pytest.fixture(scope="class")
    def audio_patch(self):
        """Return the example sin wave patch."""
        return dc.get_example_patch("sin_wav", sample_rate=500)

    @pytest.fixture(scope="class")
    def wave_dir(self, audio_patch, tmp_path_factory):
        """Create a wave directory, return path."""
        new = Path(tmp_path_factory.mktemp("wavs"))
        dc.write(audio_patch, new, "wav")
        return new

    def test_directory(self, wave_dir, audio_patch):
        """Sanity checks on wav directory."""
        assert wave_dir.exists()
        wavs = list(wave_dir.rglob("*.wav"))
        assert len(wavs) == len(audio_patch.coords.get_array("distance"))

    def test_write_single_file(self, audio_patch, tmp_path_factory):
        """Ensure a single file can be written."""
        path = tmp_path_factory.mktemp("wave_temp") / "temp.wav"
        dc.write(audio_patch, path, "wav")
        assert path.exists()

    def test_resample(self, audio_patch, tmp_path_factory):
        """Ensure resampling changes sampling rate in file."""
        path = tmp_path_factory.mktemp("wav_resample") / "resampled.wav"
        dc.write(audio_patch, path, "wav", resample_frequency=1000)
        (sr, ar) = read_wav(str(path))
        assert sr == 1000



"""
Support for prodML format.

This is used by Silixa's iDAS as an HDF5 format and perhaps other interrogators.

More info about ProdML can be found here:
https://www.energistics.org/prodml-developers-users
"""
from __future__ import annotations
from .core import ProdMLV2_0, ProdMLV2_1
